#include "FWIRKOperator.hpp"

#include "DGAssemble.hpp"
#include "Helper/FWIHelpers.hpp"
#include "lib5_solve/solver/preconditioner/Jacobi.hpp"

RKOperator::RKOperator(Matrix *M_, Matrix *A_, const Vector &U, const inteOptions &_opt)
    : M(M_), A(A_), B(U), C(U), opt(_opt) {
  B = 0.0;
  C = 0.0;

}

RKOperator::RKOperator(std::shared_ptr<DGAssemble> _assemble, const Vector &U,
                       const inteOptions &_opt) : B(U), C(U), opt(_opt),
                                                  assemble(_assemble) {
  B = 0.0;
  C = 0.0;
  assembleMatrices(*assemble, U);
}

RKOperator::RKOperator(std::shared_ptr<DGAssemble> _assemble, const Vector &U,
                       const inteOptions &_opt, int rkorder, double dt) : B(U), C(U), opt(_opt),
                                                                          assemble(_assemble) {
  B = 0.0;
  C = 0.0;
  assembleMatrices(*assemble, U);
  initialize(rkorder, dt);
}

void RKOperator::assembleMatrices(DGAssemble &assemble_, const Vector &U) {
  cleanUpMatrizes = true;
  A = std::make_unique<Matrix>(U);
  M = std::make_unique<Matrix>(U);
  assemble_.MassMatrix(*M);
  assemble_.FluxMatrix(*A);
  M->CreateSparse();
  A->CreateSparse();
}

void RKOperator::initialize(int rkorder, double _dt) {
  initialized = true;
  dt = _dt;
  fillQuadrature(c, b, rkorder);
  if (rkorder == -2) {Exit("order -2 not suitable for this operator"); }
  for (size_t stage = 0; stage < c.size(); ++stage) {
    if (rint(c[stage]) == 1) {
      B = *A;
      B *= ((-1.0) * dt * b[stage]);
    }
    if (rint(c[stage]) == 0) {
      C = *A;
      C *= ((-1.0) * dt * b[stage]);
    }
  }
  if (!opt.fwd && opt.debug)
    std::swap(B, C); // macht den RKBwd Test negativ, erzeugt aber dierichtige Adjungierte!
  C.GetData() -= (*M).GetData();
  B.GetData() += (*M).GetData();
  B.CreateSparse();
  C.CreateSparse();
}

void RKOperator::loopFwdOneStep(const Vectors &WaveFieldPoint, Vector &WaveFieldEvaluation,
                                int timeStep) const {
  if (timeStep == 0) {
    WaveFieldEvaluation += (*M) * WaveFieldPoint[0];
  } else {
    Vector Cu(0.0, WaveFieldEvaluation);
    Vector Bu(0.0, WaveFieldEvaluation);
    Vector tmp(WaveFieldEvaluation);
    double thresholdInt = threshold;
    if (static_cast<double>(timeStep) / static_cast<double>(WaveFieldEvaluation.size()) >
        threshold) {
      thresholdInt += thresholdIncrease;
    }
    if (!WaveFieldPoint[timeStep - 1].GetAccumulateFlag()) {
      tmp = WaveFieldPoint[timeStep - 1];
      tmp.Accumulate();
      Cu = C * tmp;
    } else {
      Cu = C * WaveFieldPoint[timeStep - 1];
    }
    if (!WaveFieldPoint[timeStep].GetAccumulateFlag()) {
      tmp = WaveFieldPoint[timeStep];
      tmp.Accumulate();
      Bu = B * tmp;
    } else {
      Bu = B * WaveFieldPoint[timeStep];
    }
    Bu += Cu;
    WaveFieldEvaluation += Bu;
  }

}

void RKOperator::loopBwdOneStep(const Vectors &WaveFieldPoint, Vector &WaveFieldEvaluation,
                                int timeStep) const {
  if (timeStep == WaveFieldPoint.size() -1) {
    mout << "Starting Evaluation" << endl;
    Vector u(WaveFieldPoint[timeStep]);
    if (!u.GetAccumulateFlag()) u.Accumulate();
    WaveFieldEvaluation += (*M) * u;
  } else {
    Vector Cu(0.0, WaveFieldEvaluation);
    Vector Bu(0.0, WaveFieldEvaluation);
    Vector tmp(WaveFieldEvaluation);
    double thresholdInt = threshold;
      double quotient =
          static_cast<double>(timeStep) / static_cast<double>(WaveFieldEvaluation.size());
//      if (quotient > thresholdInt) {
//        mout << "Evaluation at " << thresholdInt * 100 << "% Time needed: " << Date() - Time
//             << endl;
//        thresholdInt += thresholdIncrease;
//      }
      if (!WaveFieldPoint[timeStep + 1].GetAccumulateFlag()) {
        tmp = WaveFieldPoint[timeStep + 1];
        tmp.Accumulate();
        Cu = C * tmp;
      } else {
        Cu = C * WaveFieldPoint[timeStep + 1];
      }
      if (!WaveFieldPoint[timeStep].GetAccumulateFlag()) {
        tmp = WaveFieldPoint[timeStep];
        tmp.Accumulate();
        Bu = B * tmp;
      } else {
        Bu = B * WaveFieldPoint[timeStep];
      }
      Bu += Cu;
      WaveFieldEvaluation += Bu;
    }
}


