#include "FWINonlinearSolvers.hpp"

#include <deque>
#include <iostream>
#include <memory>
#include <valarray>

#include "ForwardModels/ForwardModel.hpp"
#include "ForwardModels/ForwardModelAAO.hpp"
#include "Helper/FWIDataContainerDomain.hpp"
#include "Helper/FWIDataContainerRange.hpp"
#include "Helper/RegularisationStrategy.hpp"
#include "Spectrum.hpp"

void FWINonlinearSolver::backtracking(
    const FWISubproblem &dataObs, std::vector<double> &residualNorms,
    const FWIDataContainerDomain &iterateMUpdate,
    FWIDataContainerDomain &iterateNonlinear,
    FWISubproblem &nonlinearEvaluation, FWISubproblem &defect,
    std::vector<int> subProb) {
  FWIDataContainerDomain iterateNonlinearTMP(iterateNonlinear);
  iterateNonlinearTMP += iterateMUpdate;
  FWD.NonlinearEvaluation(iterateNonlinearTMP, nonlinearEvaluation, subProb);
  FWD.CalculateDefect(dataObs, nonlinearEvaluation, defect, subProb);
  double r_norm = sqrt(FWD.ScalarProductL2(defect, defect));
  if (r_norm < residualNorms[iter]) {
    mout << "====================================" << endl;
    mout << "No Backtracking needed! New Residual: " << r_norm << " vs "
         << residualNorms[iter] << endl;
    iterateNonlinear = iterateNonlinearTMP;
    mout << OUT(iter) << endl;
    if (iter == 0)
      residualNorms[1] = r_norm;
    mout << OUT(residualNorms[iter]) << endl;
    return;
  }
  mout << "iterate Norm before backtracking starts: "
       << FWD.ScalarProductL2(iterateNonlinear, iterateNonlinear) << endl;
  mout << "update before backtracking : "
       << FWD.ScalarProductL2(iterateMUpdate, iterateMUpdate) << endl;
  int back = 1;
  while (r_norm >= residualNorms[iter]) {
    mout << "====================================" << endl;
    mout << "  " + Name() + " d(" << iter << "): Backtracking(" << back - 1
         << "): Starting ( " << r_norm << " vs " << residualNorms[iter] << " )";
    iterateNonlinearTMP = iterateNonlinear;
    FWIDataContainerDomain iterateMUpdateTMP(iterateMUpdate);
    double t = pow(BacktrackingFactor, back);
    iterateMUpdateTMP *= t;
    iterateNonlinearTMP += iterateMUpdateTMP;
    FWD.NonlinearEvaluation(iterateNonlinearTMP, nonlinearEvaluation, subProb);
    FWD.CalculateDefect(dataObs, nonlinearEvaluation, defect, subProb);
    r_norm = sqrt(FWD.ScalarProductL2(defect, defect));
    mout << "  " + Name() + " d(" << iter << "): Backtracking finished ( "
         << r_norm << " vs " << residualNorms[iter] << " )" << endl;
    ++back;

    mout << " the backtracking factor is: " << t << endl;
  }
  iterateNonlinear = iterateNonlinearTMP;
  if (iter == 0)
    residualNorms[1] = r_norm;
}

void FWINonlinearSolver::backtrack(const FWISubproblem &dataObs,
                                   std::vector<double> &residualNorms,
                                   const FWIDataContainerDomain &iterateMUpdate,
                                   FWIDataContainerDomain &iterateNonlinear,
                                   FWISubproblem &nonlinearEvaluation,
                                   FWISubproblem &defect,
                                   std::vector<int> subProb) {
  backtracking(dataObs, residualNorms, iterateMUpdate, iterateNonlinear,
               nonlinearEvaluation, defect, subProb);
}

class FWIREGINNBASE : public FWINonlinearSolver {
  void initializeIterateList(const FWIDataContainerDomain &iterateNonlinear,
                             std::vector<FWIDataContainerDomain> &iterateList) {
    if (FWD.GetReconstructionType() == "AllAtOnce") {
      bool initialAAO = true;
      Config::Get("AAORecalculateWFEveryOuter", initialAAO);
      if (initialAAO) {
        return;
      }
      DGSeismogram stmp(DGSeismogram(FWD.GetObservationOperator().GetSpaceOp().GetParaDat()));
      for (int i = 0; i < FWD.GetNumberOfShots(); ++i) {
        iterateList.push_back(iterateNonlinear);
        FWD.Parameter2SeismogramMap(iterateNonlinear.getMat(), stmp, i,
                                    &iterateList[i].getVectorsWrite());
      }
    }
  }
  void getIterateFromlist(const std::vector<FWIDataContainerDomain> &,
                          const std::vector<int> &prob,
                          FWIDataContainerDomain &iterateNonlinear) {
    if (prob.size() > 1) {
      Exit("just for kaczmarcz implemented!")
    }
    bool initialAAO = true;
    if (FWD.GetReconstructionType() == "AllAtOnce") {
      Config::Get("AAORecalculateWFEveryOuter", initialAAO);
      if (initialAAO) {
        DGSeismogram stmp(DGSeismogram(FWD.GetObservationOperator().GetSpaceOp().GetParaDat()));
        FWD.Parameter2SeismogramMap(iterateNonlinear.getMat(), stmp, prob[0],
                                    &iterateNonlinear.getVectorsWrite());
      }
    }
  }

  void finishIterate(vector<FWIDataContainerDomain> &iterateList,
                     const std::vector<int> &prob,
                     FWIDataContainerDomain &iterateNonlinear,
                     FWISubproblem &nonlinearEvaluations) {
    if (FWD.GetReconstructionType() == "AllAtOnce") {
      bool initialAAO = true;
      Config::Get("AAORecalculateWFEveryOuter", initialAAO);
      if (initialAAO) {
        nonlinearEvaluations[0].resetWaveData();
        return;
      } else {
        if (prob.size() > 1) {
          Exit("just for kaczmarcz implemented!")
        }
        iterateList[prob[0]].getVectorsWrite() = iterateNonlinear.getVectors();
      }
    }
  }

public:
  string nameSuff;

protected:
  std::vector<vector<int>> subproblems;
  ReginnOptions opt;
  int all_shots_iter = 0;

public:
  FWIREGINNBASE(ForwardModel &FWD, std::string name, ReginnOptions _opt)
      : FWINonlinearSolver(name.c_str(), FWD), opt(_opt) {
    verbose = 5;
    iter = 0;
  }

  ~FWIREGINNBASE() override = default;

  void initializeSubproblems(const std::vector<FWIDataContainerRange> &data_obs,
                             std::vector<FWISubproblem> &nonlinearEvaluations) {
    bool initialAAO = true;
    Config::Get("AAORecalculateWFEveryOuter", initialAAO);
    subproblems.resize(data_obs.size());
    nonlinearEvaluations.resize(data_obs.size());
    for (int i = 0; i < data_obs.size(); ++i) {
      subproblems[i].resize(1);
      nonlinearEvaluations[i].emplace_back(DGSeismogram(FWD.GetObservationOperator().GetSpaceOp().GetParaDat()),
                            Vectors(FWD.TConf.timeSteps+1,FWD.GetWaveVec()));
      if (initialAAO)
        nonlinearEvaluations[i].back().resetWaveData();
      subproblems[i][0] = i;
    }
  }

  virtual double innerLoopOneStep(
      const FWIDataContainerDomain &iterateNonlinear,
      const FWISubproblem &nonlinearEvaluation, FWISubproblem &defectLinearized,
      FWIDataContainerDomain &iterateMUpdate, int innerIter, size_t shot) = 0;

  void operator()(const IAssemble &, Vector &) override {
    Exit("not implemented");
  }

  NonlinearOutput operator()(const std::vector<FWIDataContainerRange> &dataObs,
                             FWIDataContainerDomain &iterateNonlinear,
                             std::string nameSuff_) override {
    mout << "Memory Usage at start" << PPM->TotalMemoryUsage() << endl;
    nameSuff = nameSuff_;
    Start = Date();
    vector<FWISubproblem> nonlinearEvaluations;
    vector<FWIDataContainerDomain> iterateList;
    FWISubproblem defect;
    int writeSeismoToFileEvery = INT_MAX;
    Config::Get("WriteSeismoToFileEvery", writeSeismoToFileEvery);
    initializeIterateList(iterateNonlinear, iterateList);
    initializeSubproblems(dataObs, nonlinearEvaluations);
    mout << "Initialize Subproblems " << PPM->TotalMemoryUsage() << endl;
    size_t num_probs = subproblems.size();
    FWIDataContainerDomain iterateMUpdate(FWD.GetReconstructionType(),
                                          FWD.GetMatVec(), FWD.GetWaveVec(),
                                          FWD.TConf.timeSteps);
    vector<RegularisationStrategyWinkler> regs;
    regs.resize(subproblems.size(),
                RegularisationStrategyWinkler(opt.gamma, opt.thetaMax,
                                              opt.maxInnerLoopSteps));
    d_0 = 0.0;
    std::cout << std::setprecision(15);
    size_t substeps = 1;
    size_t warmup = 0;
    Config::Get("substeps", substeps);
    Config::Get("warmup", warmup);
    for (; iter < max_iter; ++iter) {
      std::vector<int> order(num_probs);
      std::iota(order.begin(), order.end(), 0);
      //      std::random_device rd;
      //      std::mt19937 gen(42);
      //      std::shuffle(order.begin(), order.end(), gen);
      for (size_t i = 0; i < num_probs; ++i) {
        int prob = order[i];
        FWISubproblem &nonlinearEvaluation = nonlinearEvaluations[prob];
        for (size_t substep = 0; substep < substeps; ++substep) {
          mout << "before get iterate " << PPM->TotalMemoryUsage() << endl;
          if (substep == 0) getIterateFromlist(iterateList, subproblems[prob], iterateNonlinear);
          mout << "after get iterate " << PPM->TotalMemoryUsage() << endl;
          vout(2) << Name() + " problem " << prob << ": d(" << iter << ")" << endl;
          FWD.NonlinearEvaluation(iterateNonlinear, nonlinearEvaluation, subproblems[prob]);
          mout << "after NonlinearEvaluation " << PPM->TotalMemoryUsage() << endl;
          if (iter % writeSeismoToFileEvery == 0) {
            nonlinearEvaluation[0].getSeismogram().WriteToFile(Config::GetDataPath() + "FWI/" +
                                                               "Seis" + std::to_string(prob) + "_" +
                                                               std::to_string(iter));
          }
          FWD.CalculateDefect(dataObs, nonlinearEvaluation, defect, subproblems[prob]);
          if (FWD.GetReconstructionType() == "AllAtOnce" && substep == 0) {
            defect[0].getVectorsWrite().Clear();
          }
          mout << "after Defect " << PPM->TotalMemoryUsage() << endl;
          if (defect.size() != nonlinearEvaluation.size()) {
            Exit(" DataContainer sizes do not match")
          }
          if (iter != 1) regs[prob].AddResNorm(sqrt(FWD.ScalarProductL2(defect, defect)));
          regs[prob].CalcTheta();
          vout(1) << "*************************************" << endl
                  << "  " + Name() + " Subproblem " << prob << ": d(" << iter
                  << "): Starting inner loop " << "( ||r||_2 = " << regs[prob].GetResNorm(iter)
                  << ", theta*||r||_2 = " << regs[prob].CalcInnerStopp()
                  << ", theta = " << regs[prob].GetTheta(iter)
                  << ", m_max = " << regs[prob].CalcMaxInner() << " )" << endl;
          int innerIter = 0;
          iterateMUpdate.Clear();
          double residualNormTilde = infty;
          while ((innerIter < min(static_cast<unsigned int>(opt.maxInnerLoopSteps),
                                  regs[prob].CalcMaxInner()) &&
                  residualNormTilde > regs[prob].CalcInnerStopp()) ||
                 innerIter < min_iter) {
            mout << "Before innerstep " << PPM->TotalMemoryUsage() << endl;
            MOUT(FWD.ScalarProductL2(iterateMUpdate, iterateMUpdate));
            residualNormTilde = innerLoopOneStep(iterateNonlinear, nonlinearEvaluation, defect,
                                                 iterateMUpdate, innerIter, prob);
            MOUT(FWD.ScalarProductL2(iterateMUpdate, iterateMUpdate));
            if (substep < warmup) { iterateMUpdate.getMatWrite().Clear(); }
            mout << "after innerstep " << PPM->TotalMemoryUsage() << endl;
            vout(2) << "  " + Name() + " Subproblem: " << prob << ": d(" << iter << "), inner("
                    << innerIter << "): " << "||r||_2 = " << residualNormTilde << endl;
            innerIter++;
          }
          if (iter == 0) regs[prob].AddResNorm(residualNormTilde);
          regs[prob].AddInnerIter(innerIter);

          vout(1) << "  " + Name() + " Subproblem: " << prob << ": d(" << iter
                  << "): Finished inner loop " << "( " << regs[prob].GetInnerIter(iter)
                  << " steps )" << endl
                  << "*************************************" << endl;
          if (FWD.UseBoxConstraint()) {
            FWD.ApplyBoxConstraint(iterateNonlinear, iterateMUpdate);
          } else {
            iterateNonlinear += iterateMUpdate;
          }
          if (FWD.UseGardener()) { FWD.Gardener(iterateNonlinear); }
          if (BacktrackingActivated) {
            if (FWD.GetReconstructionType() == "AllAtOnce") { Exit("Defect is set to zero in AAO") }
            backtrack(dataObs, regs[prob].GetResNormVector(iter), iterateMUpdate, iterateNonlinear,
                      nonlinearEvaluation, defect, subproblems[prob]);
          }
          int plotIter = 1;
          Config::Get("PlotFactorIterates", plotIter);
          int plotShot = -1;
          Config::Get("PlotShot", plotShot);

          if (iter % plotIter == 0 && (plotShot < 0 || prob == plotShot)) {
            FWD.plot(iterateNonlinear,
                     "_fwi_" + Name() + nameSuff + "_iterate_m_" + to_string(prob) + "_" +
                         to_string(iter),
                     -1);
            FWD.plot(iterateNonlinear, "_fwi_" + Name() + nameSuff + "_iterate_m_" + "_",
                     all_shots_iter);
          }
          ++all_shots_iter;
        }
        mout << "Before finishIterate " << PPM->TotalMemoryUsage() << endl;
        finishIterate(iterateList, subproblems[prob], iterateNonlinear, nonlinearEvaluation);

        mout << "After finishIterate " << PPM->TotalMemoryUsage() << endl;
        if (iter + 1 == max_iter) {
          mout << "Calculating Final Residual: " << endl;
          FWD.NonlinearEvaluation(iterateNonlinear, nonlinearEvaluation, subproblems[prob]);

          FWD.CalculateDefect(dataObs, nonlinearEvaluation, defect, subproblems[prob]);
          if (FWD.GetReconstructionType() == "AllAtOnce") { defect[0].getVectorsWrite().Clear(); }
          regs[prob].AddResNorm(sqrt(FWD.ScalarProductL2(defect, defect)));
        }
      }
    }
    vout(1) << " " + Name() + ": d(" << iter << ")= " << d << "\n";

    NonlinearOutput output;
    for (size_t prob = 0; prob < num_probs; ++prob) {
      std::string shotStr = "NonlinearResiduals Shot " + to_string(prob);
      mout.printEntryRaw(shotStr, regs[prob].GetResNormVector(0));
      std::string nrOfInnerStr = "Inner Iterations Shot " + to_string(prob);
      mout.printEntryRaw(nrOfInnerStr, regs[prob].GetInnerIters());
      output.push_back(regs[prob].GetResNormVector(0));
    }
    return output;
  }
};

class FWICGREGINN_ : public FWIREGINNBASE {
  FWIDataContainerDomain iterateUpdatePN;
  FWIDataContainerDomain iterateUpdatePNMinus1;
  double beta = 0.0;

public:
  virtual auto initializePreconditioner(
      const FWIDataContainerDomain &nonlinearPoint,
      const FWISubproblem &nonlinearEvaluation,
      const FWIDataContainerDomain &mNegativeUpdate,
      FWISubproblem &defectLinearized, int innerIter) -> void = 0;

  virtual void applyPreconditioner(const FWIDataContainerDomain &,
                                   FWIDataContainerDomain &, FWISubproblem &,
                                   int) = 0;

  FWICGREGINN_() = delete;

  FWICGREGINN_(ForwardModel &FWD, std::string name, ReginnOptions opt)
      : FWIREGINNBASE(FWD, std::move(name), opt),
        iterateUpdatePN(FWD.GetReconstructionType(), FWD.GetMatVec(),
                        FWD.GetWaveVec(), FWD.TConf.timeSteps),
        iterateUpdatePNMinus1(FWD.GetReconstructionType(), FWD.GetMatVec(),
                              FWD.GetWaveVec(), FWD.TConf.timeSteps) {}

  double innerLoopOneStep(const FWIDataContainerDomain &iterateNonlinear,
                          const FWISubproblem &nonlinearEvaluation,
                          FWISubproblem &defectLinearized,
                          FWIDataContainerDomain &iterateMUpdate, int innerIter,
                          size_t prob) override {
    FWISubproblem iterateLinearizedTMP(defectLinearized);
    if (innerIter == 0)
      beta = 0.0;
    vout(2) << "  " + Name() + " Subproblem " << prob << ": d(" << iter
            << "), inner(" << innerIter << "): "
            << "Solving backpropagation problem" << endl;
    iterateUpdatePN.Clear();
    MOUT(FWD.ScalarProductL2(iterateUpdatePN, iterateUpdatePN));
    FWD.AdjointLinearized(iterateNonlinear, defectLinearized, iterateUpdatePN);
    if (fwiverbose >= 2)
      FWD.plot(iterateUpdatePN,
               "_fwi_" + Name() + nameSuff + "Gradient_OUTER_" +
                   to_string(iter) + "_INNER_" + to_string(innerIter),
               -1);
    MOUT(FWD.ScalarProductL2(iterateUpdatePN, iterateUpdatePN));
    initializePreconditioner(iterateNonlinear, nonlinearEvaluation,
                             iterateUpdatePN, defectLinearized, innerIter);
    FWIDataContainerDomain iterateUpdatePNPC(iterateUpdatePN);
    applyPreconditioner(iterateNonlinear, iterateUpdatePNPC, defectLinearized,
                        innerIter);
    MOUT(FWD.ScalarProductL2(iterateUpdatePN, iterateUpdatePN));

    double skpPCd = FWD.ScalarProductL2(iterateUpdatePNPC, iterateUpdatePN);
    iterateUpdatePN = iterateUpdatePNPC;
    MOUT(FWD.ScalarProductL2(iterateUpdatePN, iterateUpdatePN));
    iterateUpdatePNMinus1 *= (beta * skpPCd);
    beta = 1.0 / skpPCd;
    iterateUpdatePN += iterateUpdatePNMinus1;
    vout(2) << "  " + Name() << " Subproblem" << prob << ": d(" << iter
            << "), inner(" << innerIter << "):  "
            << "Solving linearized forward problem" << endl;
    FWD.Linearized(iterateNonlinear, iterateUpdatePN, iterateLinearizedTMP);
    double linearizedNorm =
        FWD.ScalarProductL2(iterateLinearizedTMP, iterateLinearizedTMP);
    double alpha = skpPCd / linearizedNorm;
    MOUT(alpha);
    iterateUpdatePN *= alpha;
    iterateMUpdate += iterateUpdatePN;
    iterateUpdatePN *= 1.0 / alpha;
    iterateLinearizedTMP *= alpha;
    defectLinearized -= iterateLinearizedTMP;
    iterateUpdatePNMinus1 = iterateUpdatePN;
    if (fwiverbose >= 2)
      FWD.plot(iterateUpdatePN,
               "_fwi_" + Name() + nameSuff + "Iter_OUTER_" + to_string(iter) +
                   "_INNER_" + to_string(innerIter),
               -1);
    return sqrt(FWD.ScalarProductL2(defectLinearized, defectLinearized));
  }
};
class FWICGREGINNCanonic : public FWICGREGINN_ {
public:
  auto initializePreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                                const FWISubproblem &nonlinearEvaluation,
                                const FWIDataContainerDomain &mNegativeUpdate,
                                FWISubproblem &defectLinearized, int innerIter)
      -> void override {
    FWD.InitializePrecon(nonlinearPoint.getMat(), mNegativeUpdate.getMat(),
                         nonlinearEvaluation[0].getVectors());
  }

  auto applyPreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                           FWIDataContainerDomain &mNegativeUpdate,
                           FWISubproblem &defectLinearized, int innerIter)
      -> void override {
    FWD.ApplyPrecon(nonlinearPoint, mNegativeUpdate);
  }

  FWICGREGINNCanonic() = delete;

  FWICGREGINNCanonic(ForwardModel &FWD, std::string name_, ReginnOptions opt)
      : FWICGREGINN_(FWD, name_, opt) {}
};

class FWISDREGINN : public FWIREGINNBASE {
protected:
  FWIDataContainerDomain iterateUpdatePN;

public:
  FWISDREGINN(ForwardModel &FWD, std::string name_, ReginnOptions opt)
      : FWIREGINNBASE(FWD, name_, opt),
        iterateUpdatePN(FWD.GetReconstructionType(), FWD.GetMatVec(),
                        FWD.GetWaveVec(), FWD.TConf.timeSteps) {}

  virtual void multiplyWithSteplength(const FWIDataContainerDomain &,
                                      FWIDataContainerDomain &,
                                      FWIDataContainerDomain &,
                                      FWISubproblem &) = 0;

  double innerLoopOneStep(const FWIDataContainerDomain &iterateNonlinear,
                          const FWISubproblem &nonlinearEvaluation,
                          FWISubproblem &defectLinearized,
                          FWIDataContainerDomain &iterateMUpdate, int innerIter,
                          size_t shot) override {
    vout(2) << "  " + Name() + " shot" << shot << ": d(" << iter << "), inner("
            << innerIter << "): "
            << "Solving backpropagation problem" << endl;
    iterateUpdatePN.Clear();
    std::cout << std::setprecision(15);
    FWD.AdjointLinearized(iterateNonlinear, defectLinearized, iterateUpdatePN);
    multiplyWithSteplength(iterateNonlinear, iterateUpdatePN, iterateMUpdate,
                           defectLinearized);
    return sqrt(FWD.ScalarProductL2(defectLinearized, defectLinearized));
  }

  ~FWISDREGINN() override = default;
};
class FWISDREGINNMultiAAO : public FWISDREGINN {
  ForwardModelAAO *FWDAAO;

public:
  FWISDREGINNMultiAAO(ForwardModel &FWD, std::string name, ReginnOptions opt)
      : FWISDREGINN(FWD, std::move(name), opt) {
    FWDAAO = dynamic_cast<ForwardModelAAO *>(&FWD);
  }
  ~FWISDREGINNMultiAAO() override = default;

  void multiplyWithSteplength(const FWIDataContainerDomain &iterateNonlinear,
                              FWIDataContainerDomain &iterateUpdatePN,
                              FWIDataContainerDomain &iterateMUpdate,
                              FWISubproblem &defectLinearized) override {
    FWD.InitializePrecon(iterateNonlinear.getMat(), iterateUpdatePN.getMat(),
                         iterateNonlinear.getVectors());
    FWD.ApplyPrecon(iterateNonlinear, iterateUpdatePN);
    FWDAAO->MSDProcedure(iterateNonlinear, iterateUpdatePN, iterateMUpdate, defectLinearized);
    FWISubproblem linearizedCalcs(defectLinearized);
    FWD.Linearized(iterateNonlinear, iterateUpdatePN, linearizedCalcs);
    defectLinearized -= linearizedCalcs;
    iterateMUpdate += iterateUpdatePN;
  }
};

  class FWISDREGINNMulti : public FWISDREGINN {
    std::vector<unsigned int> InvMap;

  public:
    FWISDREGINNMulti(ForwardModel &FWD, std::string name, ReginnOptions opt)
        : FWISDREGINN(FWD, std::move(name), opt) {
      InvMap = FWD.GetReconstructionVector();
    }

    ~FWISDREGINNMulti() override = default;

    void applyPreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                             FWIDataContainerDomain &mNegativeUpdate) {
      FWD.ApplyPrecon(nonlinearPoint, mNegativeUpdate);
    }

    void multiplyWithSteplength(const FWIDataContainerDomain &iterateNonlinear,
                                FWIDataContainerDomain &iterateUpdatePN,
                                FWIDataContainerDomain &iterateMUpdate,
                                FWISubproblem &defectLinearized) override {
      FWIDataContainerDomain mTemp(iterateUpdatePN);
      int vecsize = InvMap.size();
      std::vector<FWISubproblem> linearizedCalcs(vecsize, defectLinearized);
      for (size_t i = 0; i < InvMap.size(); ++i) {
        mTemp = iterateUpdatePN;
        for (size_t j = 0; j < FWD.GetParameterComponents(); ++j) {
          if (InvMap[i] != j)
            mTemp.setCompToZero(j);
        }
        FWD.Linearized(iterateNonlinear, mTemp, linearizedCalcs[i]);
      }
      int matsize = FWD.GetParameterComponents();
      SymRMatrix matrixEntries2(matsize);
      matrixEntries2.Identity();
      for (size_t i = 0; i < InvMap.size(); ++i) {
        size_t pos1 = InvMap[i];
        for (size_t j = 0; j < InvMap.size(); ++j) {
          size_t pos2 = InvMap[j];
          matrixEntries2(FWD.ScalarProductL2(linearizedCalcs[i],
                                             linearizedCalcs[j]),
                         pos1, pos2);
        }
      }
      RVector rightSideOne(0.0, matsize);
      RVector rightSideTwo(0.0, matsize);
      for (size_t i = 0; i < InvMap.size(); ++i) {
        rightSideOne[InvMap[i]] =
            FWD.ScalarProductL2(defectLinearized, linearizedCalcs[i]);
        mTemp = iterateUpdatePN;
        for (size_t j = 0; j < FWD.GetParameterComponents(); ++j) {
          if (j != InvMap[i])
            mTemp.setCompToZero(j);
        }
        rightSideTwo[InvMap[i]] = FWD.ScalarProductL2(mTemp, iterateUpdatePN);
      }
      MOUT(rightSideOne);
      MOUT(rightSideTwo);
      RVector calculatedFactors(0.0, matsize);
      SymRMatrix matrixEntries3(matsize);
      matrixEntries3.Identity();
      matrixEntries3 *= 10e-16;
      matrixEntries2 += matrixEntries3;
      MOUT(matrixEntries2);
      matrixEntries2.Invert();
      calculatedFactors = matrixEntries2 * rightSideOne;
      MOUT(rightSideOne);
      MOUT(calculatedFactors);
      vout(4) << "the ts are: ";
      for (int i = 0; i < InvMap.size(); ++i) {
        std::string a = " t_" + std::to_string(i) + " = " +
                        std::to_string(calculatedFactors[i]);
        vout(4) << a;
      }

      vout(4) << endl;
      for (size_t i = 0; i < InvMap.size(); ++i) {
        linearizedCalcs[i] *= calculatedFactors[InvMap[i]];
        defectLinearized -= linearizedCalcs[i];
      }

      mTemp = iterateUpdatePN;
      mTemp *= calculatedFactors;

      iterateMUpdate += mTemp;
      mout << "norm of the new iterate: "
           << OUT(FWD.ScalarProductL2(iterateMUpdate, iterateMUpdate)) << endl;
    }
  };

  FWINonlinearSolver *GetFWINonlinearSolver(ForwardModel &FWD) {
    std::string solver = "FWISteepestDescent";
    Config::Get("FWINonlinearSolver", solver);
    if (solver == "FWI_CGREGINN") {
      std::string name = "CGREGINN";
      std::string type = "Canonic";
      Config::Get("CGREGINNType", type);
      ReginnOptions opt(name);
      name = name + type;
      if (type == "Canonic") {
        return new FWICGREGINNCanonic(FWD, name, opt);
      }
    } else if (solver == "FWI_mSDREGINN" && FWD.GetReconstructionType() == "Classical") {
      ReginnOptions opt("mSDREGINN");
      return new FWISDREGINNMulti(FWD, "mSDREGINN", opt);
    }else if (solver == "FWI_mSDREGINN"&& FWD.GetReconstructionType() == "AllAtOnce") {
      ReginnOptions opt("mSDREGINN");
      return new FWISDREGINNMultiAAO(FWD, "mSDREGINN", opt);
    }
    Exit("Not implemented");
  }
