#include <cmath>
#include <iomanip>
#include <iostream>
#include <memory>
#include <string>
#include <utility>
#include <vector>

#include "Cell.hpp"
#include "Config.hpp"
#include "DGVAcousticAssembleCombined.hpp"
#include "DGVAcousticMatrixAssembler.hpp"
#include "DGVAcousticPlotter.hpp"
#include "DGVAcousticRHSAssembler.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "FWINonlinearSolvers.hpp"
#include "FWIProblemWrapper.hpp"
#include "FWITimeSolver.hpp"
#include "ForwardModels/ForwardModelAAO.hpp"
#include "ForwardModels/ForwardModelCreator.hpp"
#include "GlobalDefinitions.hpp"
#include "Helper/ABCHandler.hpp"
#include "Helper/FWIDataContainerDomain.hpp"
#include "Helper/FWIDataContainerRange.hpp"
#include "Helper/FWIHelpers.hpp"
#include "Helper/ReconstructionContainer.hpp"
#include "Helper/VTUModelReader.hpp"
#include "Helper/VectorFillFunctions.hpp"
#include "LoggingBackends.hpp"
#include "Logging.hpp"
#include "Mesh.hpp"
#include "Meshes.hpp"
#include "Observation/DGSeismogram.hpp"
#include "Observation/FWIObservationOperatorDGTS.hpp"
#include "ObservationSpecification.hpp"
#include "Problems/FWIAcousticInversionProb.hpp"
#include "Problems/FWIInversionProblems.hpp"
#include "Problems/Sources/FWISourceWavelet.hpp"
#include "Reduced/FWILinOpAdjRed.hpp"
#include "VAcousticMatFromVec.hpp"
#include "Vector.hpp"

class MeasureConf {
public:
  std::string measurement_time = "point";
  std::string measurement_space = "point";
  std::string ReceiverPlacement = "Waypoints";
};
void FWIInverseMainDGTSVANEWNEW() {
  auto meshes = getFWIMesh();
  WPConf Conf;
  auto prob = GetInversionProblem(Conf);
  LevelPair Lev = {0, -1};
  int deg = 0;
  FWIDiscretizationStack discStack(meshes, prob->GetParamComps(),
                                   prob->GetStateComponents(), deg, Lev);
  auto eleCache =
      std::make_unique<EleCacheContainer>(*discStack.wV, prob->GetNumL());

  std::shared_ptr<ObservationSpecification> spec = ObservationSpecificationBuilder().CreateShared();

  auto mask = GetMeasureMask(meshes->dim(), prob->GetNumL(), 0);
  std::shared_ptr<FWIObservationOperatorDGTS> Obs =
      ObservationCreator(eleCache->Get(), discStack.GetMesh(), prob->GetNumL())
          .CreateShared();
  //  std::unique_ptr<FWIForwardModelDGTSVA> FWD(getFWD(meshes, Conf.level, deg,
  //  GetInversionProblem(Conf)));

  //    std::string seismopath = "";
  //    Config::Get("seismopath", seismopath);
  //    if (seismopath == "") {
  //        FWIDataContainerDomain real(*FWD);
  //        Vector M_real(FWD->GetMatVec());
  //        M_real.Clear();
  //        FWD->fillVectorFromProblem(M_real, FWD->GetProblem());
  //        real.getMatWrite() = M_real;
  //        FWD->plot(M_real, "fwi_m_real");
  //        TIMEFUNCTION(FWD->createRealData(real, obs);)
  //        if (Conf.SaveObservedSeismogram) {
  //            FWISeismogram printObs;
  //            for (int i = 0; i < obs.size(); ++i) {
  //                printObs = obs[i].getSeismogram();
  //                printObs.collect();
  //                printObs.writeToFile(Config::getDataPath() +
  //                string("FWI/Obs") + std::to_string(i));
  //            }
  //        }
  //    } else {
  //        int shots = FWD->GetProblem().getShotCount();
  //        for (int shot = 0; shot < shots; ++shot) {
  //            std::string seismostring = seismopath + std::to_string(shot);
  //            mout << "Reading data from " + seismostring + " make sure it has
  //            " +
  //                    std::to_string(FWD->GetProblem().GetParamComps()) + "
  //                    entries!" << endl;
  //
  //        }
  //    }
}

std::shared_ptr<ForwardModel> getFWDshared(
    const std::shared_ptr<Meshes> &M, int level, int deg,
    const std::shared_ptr<FWIVAcousticInversionProb> &prob) {
  std::string type;
  Config::Get("ReconstructionType", type);
  LevelPair lev{level, -1};
  auto stack =
      std::make_shared<FWIDiscretizationStack>(M, prob->GetParamComps(),
                                               prob->GetStateComponents(), deg,
                                               lev);
  (*stack->wV).PrintInfo();
  (stack->GetMesh()).PrintInfo();
  if (type == "Classical") {
    return ForwardModelRedCreator(prob, stack).CreateShared();
  } else if (type == "AllAtOnce") {
    return ForwardModelAAOCreator(prob, stack).CreateShared();
  //    return std::make_shared<ForwardModelAAO>(M, lev, deg, prob);
  } else {
    Exit("specify type or specify a known type");
  }
}

void createMeasuredSeismogram(const std::shared_ptr<ForwardModel>& FWD,WPConf &Conf,
                              std::vector<FWIDataContainerRange> &obs) {
  std::string seismopath;
  Config::Get("seismopath", seismopath);
  if (seismopath.empty()) {
    FWIDataContainerDomain real(FWD->GetReconstructionType(), FWD->GetMatVec(),
                                FWD->GetWaveVec(), FWD->TConf.timeSteps);
    Vector M_real(FWD->GetMatVec());
    M_real.Clear();
    FillVectorFromProblem(FWD->GetProblem(), M_real);
    real.getMatWrite() = M_real;
    FWD->plot(M_real, "fwi_m_real", -1);
    double signalToNoise = -100000;
    Config::Get("SignalToNoise", signalToNoise);
    int signalToNoiseSeed = 42;
    Config::Get("SignalToNoiseSeed", signalToNoiseSeed);
    TIMEFUNCTION(FWD->createRealData(real, obs);)
    if (Conf.SConf.SaveObservedSeismogram) {
      for (int i = 0; i < obs.size(); ++i) {
        obs[i].getSeismogram().WriteToFile(Config::GetDataPath() + string("FWI/Obs")
                                                   + std::to_string(i));
      }
    }
    if (signalToNoise > -100000 + Eps) {
      int seisnr = 0;
      for (auto &seis : obs) {
        seis.getSeismogramWrite().addRelativeNoise(signalToNoise,
                                                   signalToNoiseSeed +
                                                       seisnr++ * 1000000);
      }
      if (Conf.SConf.SaveObservedSeismogram) {
        for (int i = 0; i < obs.size(); ++i) {
obs[i].getSeismogram().WriteToFile(Config::GetDataPath() +
string("FWI/ObsNoisy") + std::to_string(i));
        }
      }
    }
  } else {
    string format = "FWI";
    Config::Get("ReadFormat", format);
    int shots = FWD->GetProblem().GetShotCount();
    string expfn{};
    Config::Get("ADDAXExperimentFile", expfn);
    const auto &spaceOp = FWD->GetObservationOperator().GetSpaceOp();
    if (format == "FWI" && expfn.empty()) {
      for (int shot = 0; shot < shots; ++shot) {
        std::string seismostring = seismopath + std::to_string(shot);
        mout << "Reading data from " + seismostring + " make sure it has " +
                    std::to_string(FWD->GetProblem().GetParamComps()) +
                    " entries!"
             << endl;
        obs.emplace_back(DGSeismogram(spaceOp.GetParaDat()));
        obs[shot].getSeismogramWrite().ReadFromFile(seismostring);
      }
    } else if (format == "FWI" && !expfn.empty()) {
      json j;
      std::ifstream ifs(expfn + ".json");
      if (!ifs.good()) Exit("Failed to open file" + expfn + ".json");
      nlohmann::json Seismo = nlohmann::json::parse(ifs);
      std::string s{};
      for (const auto &lel : Seismo.at("OutputDataPath")) {
        s = lel;
      }
      s = Config::GetBuildPath() + "/" + s;
      for (std::string seismofn : Seismo.at("DataNames")) {
        obs.emplace_back(DGSeismogram(spaceOp.GetParaDat()));
        MOUT("Read " + s + seismofn + ".json")
        obs.back().getSeismogramWrite().ReadFromFile(s + seismofn);
      }
    } else if (format == "IFOS") {
      Exit("Not Supported for now!")
    }
  }
}


void fillInitialModel(ForwardModel &FWD, FWIDataContainerDomain &initial) {
  bool loadInitialModelFromFile = false;
  bool loadRealModelFromFile = false;
  Config::Get("loadInitialModelFromFile", loadInitialModelFromFile);
  Config::Get("loadRealModelFromFile", loadRealModelFromFile);
  Vector M(FWD.GetMatVec());
  M.Clear();
  FWD.GetProblemWrite().setInitialValueFlag(true);
  FillVectorFromProblem(FWD.GetProblem(), M);
  if (FWD.GetProblem().GetDim() == 2) {
    FWD.GetProblemWrite().setInitialValueFlag(false);
    FWD.fillSourceAndReceiverCellsFromProblem(M, FWD.GetProblem());
    FWD.GetProblemWrite().setInitialValueFlag(true);
  }

  if (loadInitialModelFromFile) {
    std::string type("vtu");
    Config::Get("InitialModelType", type);

    Vector Tmp(M);
    Tmp.Clear();
    std::string fn;
    Config::Get("InitialModelFileName", fn);
    if (type == "JSON"){
      ReadJSONToVec(fn+"_Rho",Tmp,0);
      ReadJSONToVec(fn+"_Vp",Tmp,1);
      if(FWD.GetProblem().GetNumL() > 0 ){  ReadJSONToVec(fn+"_Vp",Tmp,2);}

    } else {
      //      THROW("Only use JSON Interface")
      std::vector<std::string> fnv;
      if (fn.empty()) Exit("Set Initial Model Filename!") fnv.push_back(fn);
      VTUModelReader reader(fnv, M, "Vp");
      if (std::string::npos == fn.find("Exp")) {
        std::string s = "Vp";
        reader.getData(Tmp, s, 1);
      } else {
        std::string s = "VpExp";
        reader.getData(Tmp, s, 1);
      }
    }
    if(FWD.GetProblem().getParametrization() == "taup"){
      M = Tmp;
    }
    if (FWD.GetProblem().getParametrization() == "taup_positive") {
      if (type == "JSON") {
        for (cell c = M.cells(); c != M.cells_end(); ++c) {
          double tmp0 = Tmp(c(), 0);
          double tmp1 = Tmp(c(), 1);
          M(c(), 0) = tmp0 > Eps ?  log(tmp0) : 0;
          M(c(), 1) = tmp1 > Eps ?  log(tmp1) : 0;
          if(FWD.GetProblem().GetNumL() > 0){
            double tmp2 = Tmp(c(), 2);
            M(c(), 2) = tmp2> Eps ? log(tmp2) : 0;
          }

        }
      }else { THROW("Do not use this anymore!") }
    }
    }
  if (FWD.GetReconstructionType() == "AllAtOnce") {
    auto &FWDAAO = dynamic_cast<ForwardModelAAO &>(FWD);
    MOUT(M.norm());
    if (FWDAAO.ABCHAAO.get()) { FWDAAO.ABCHAAO->SetABCZero(M); }
    MOUT(M.norm());
  }
  FWD.plot(M, "fwi_m_initial", -1);
  initial.getMatWrite() = M;
}

void processInitialValue(ForwardModel &FWD,
                         std::shared_ptr<FWIDataContainerDomain> &initial,
                         const WPConf &Conf) {
  auto initialnew =
      std::make_shared<FWIDataContainerDomain>(FWD.GetReconstructionType(),
                                               FWD.GetMatVec(),
                                               FWD.GetWaveVec(),
                                               FWD.TConf.timeSteps);
  ProjectModel(initial->getMat(), initialnew->getMatWrite(),
               FWD.GetParameterComponents());
  initial = initialnew;
  SmoothMaterial(initial->getMatWrite(), FWD.GetParameterComponents(),
                 Conf.SConf.MaterialSmoothPasses);
}

vector<NonlinearOutput> FWIInverseMainDGTSVA() {
  auto meshes = getFWIMesh();
  WPConf Conf;
  int deg = 0;
  Config::Get("deg", deg);
  //  LevelPair Lev = {Conf.level, -1};

  //  FWIDiscretizationStack discStack(meshes, prob->GetParamComps(),
  //                                   prob->GetStateComponents(), deg,
  //                                   Lev);
  std::vector<NonlinearOutput> Output;
  bool firstPass = true;
  std::shared_ptr<FWIDataContainerDomain> initial = nullptr;
  std::vector<FWIDataContainerRange> obs;
  auto prob = GetInversionProblem(Conf);
  auto FWD = getFWDshared(meshes, Conf.level, deg, prob);
  createMeasuredSeismogram(FWD, Conf, obs);
  std::vector<DGSeismogram> ObsSeismos;
  if (Conf.SConf.FrequencyFiltering) {
    for (const auto &ob : obs) {
      ObsSeismos.push_back(ob.getSeismogram());
    }
  }
  for (int j = 0; j < Conf.SConf.FrequencyLevels.size(); ++j) {
    Conf.SConf.CurrentLevel = j;
    auto freq = Conf.SConf.GetFreq();
    mout << "Frequency Filter to: " << freq << endl;
    if (Conf.SConf.FrequencyFiltering) {
      for (int i = 0; i < obs.size(); ++i) {
        obs[i].getSeismogramWrite() = ObsSeismos[i];
        if (Conf.SConf.FilteringType == "LowPass") {
          obs[i].getSeismogramWrite().LowPassFilter(freq[0],
                                                    Conf.SConf.FilteringOrder);
        } else if (Conf.SConf.FilteringType == "BandPass") {
          obs[i].getSeismogramWrite().BandPassFilterIFOS(freq[0], freq[1]);
        }
        if (Conf.SConf.SaveFilteredSeismogram) {
          string s;
          for (double d : freq) {
            s += std::to_string(d);
          }
          obs[i].getSeismogram().WriteToFile(Config::GetDataPath() +
                    string("FWI/ObsFilteredShot") + to_string(i)  +
                                            Conf.SConf.FilteringType + s);
        }
      }
      if (!Conf.SConf.SrcInversion) {
        FWD->GetProblemWrite().SetTimeSource(
            GetSourceWavelet(Conf.TConf, Conf.SConf));
      }
    }
    if (!firstPass) {
      processInitialValue(*FWD, initial, Conf);
    }
    if (Conf.SConf.FrequencyFiltering) {
      FWD->SetPreconRegularization(Conf.SConf.PreconRegu[j]);
    }

    std::cout << std::setprecision(15);
    if (firstPass) {
      initial =
          std::make_shared<FWIDataContainerDomain>(FWD->GetReconstructionType(),
                                                   FWD->GetMatVec(),
                                                   FWD->GetWaveVec(),
                                                   FWD->TConf.timeSteps);
      fillInitialModel(*FWD, *initial);
      firstPass = false;
    }
    if (Conf.SConf.SrcInversion) {
      std::vector<FWIDataContainerRange> obsTmp;
      FWD->createRealData(*initial, obsTmp);
      SourceInversionProblem(obs, obsTmp, FWD->problem_fwd,
                             FWD->GetObservationOperator(),
                             Conf.SConf.SrcInversionEps,
                             Conf.SConf.SrcInversionKappa, j);
    }
    std::unique_ptr<FWINonlinearSolver> NLS(GetFWINonlinearSolver(*FWD));
    std::string suffix;
    if (Conf.SConf.FrequencyFiltering) {
      for (double d : freq) {
        suffix += std::to_string((int)(d * 100));
      }
    }
    Output.push_back((*NLS)(obs, *initial, suffix));
    if (Conf.SConf.FrequencyFiltering) {
      mout << "Loop over" + suffix + "ended" << endl;
      mout << "The residuals are:" << endl;
      const NonlinearOutput &CurrentOutput = Output.back();
      for (int shot = 0; shot < CurrentOutput.size(); ++shot) {
        mout << "Residuals of shot " + to_string(shot) << endl;
        for (const auto &Result : CurrentOutput[shot]) {
          mout << Result << endl;
        }
      }
    }
  }
  return Output;
}

int createKernels2() {
  auto meshes = getFWIMesh();
  WPConf Conf;
  LevelPair lev{Conf.level, -1};
  int deg = 0;
  int numL = 0;
  Config::Get("numL", numL);

  Config::Get("deg", deg);
  Conf.s_prob = "VAcousticWaveC2Benchmark";
  std::shared_ptr<FWIVAcousticInversionProb> problem = GetInversionProblem(Conf);
  FWIDiscretizationStack Stack(meshes, problem->GetParamComps(),
                               problem->GetStateComps(), deg, lev);
  (*Stack.wD)().PrintInfo();
  auto eleCache =
      std::make_unique<EleCacheContainer>(*Stack.wV, problem->GetNumL());
  auto spec = ObservationSpecificationBuilder()
                  .WithT(Conf.TConf.T)
                  .Withdt(Conf.TConf.dt)
                  .Witht0(Conf.TConf.t0)
                  .CreateShared();
  FWIObservationOperatorDGTS *Obs =
      ObservationCreator(eleCache->Get(), Stack.GetMesh(), problem->GetNumL())
          .WithSpec(spec)
          .Create();
  std::cout << std::setprecision(15);
  problem->setInitialValueFlag(true);
  auto M = std::make_unique<Vector>(*Stack.mV);
  auto M2 = std::make_unique<Vector>(*Stack.mV);
  M2->Clear();
  M->Clear();
  FillVectorFromProblem(*problem, *M2);
  ReparametrizeMaterialVector(*M2, *problem, *M);
  auto PMLH = std::make_shared<ABCHandler>();

  problem->SetShot(0);
  FWITimeSolver TSolveR(*Stack.wV, PMLH, Conf.TConf.rkorder);
  auto Mat = std::make_unique<VAcousticMatFromVec>(*problem, *M2);
  auto Mat2 = std::make_unique<VAcousticMatFromVec>(*problem, *M2);
  auto MatAss =
      std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), true,
                                                   eleCache->Get().get());
  auto plotter = std::make_shared<DGVAcousticPlotter>(problem->GetNumL(),
                                                      eleCache->Get().get());
  std::shared_ptr<FWIProblemWrapper> problemVA =
      std::make_shared<FWIProblemWrapper>(*problem, *problem, std::move(M),
                                          problem->GetNumL(), problem->GetDim(),
                                          problem);

  auto rhsAssembler =
      std::make_shared<DGVAcousticRHSAssemblerDefault>(*problemVA);
  auto assembleCombined =
      std::make_shared<DGVAcousticAssembleCombined>(plotter, MatAss, rhsAssembler, Conf.TConf.t0,
                                                    Conf.TConf.T, Conf.TConf.t0);
  Vectors U_m(Conf.TConf.timeSteps + 1, *Stack.wV);
  U_m.Clear();
  assembleCombined->SetEleCache(eleCache->Get().get());
  assembleCombined->SetInitialValue(U_m[0]);
  TSolveR.Solve(assembleCombined, U_m);
  ReconstructionContainer recon(*problem);
  recon.Vp = true;
  recon.Rho = true;
  recon.Taup = true;
  FWILinOpAdjRed AdjOp(U_m, *M2, *Obs, problem, PMLH, Conf.TConf, recon,
                       eleCache->Get());
  DGSeismogram Seis(Obs->GetSpaceOp().GetParaDat());
  Vector Eval(*Stack.mV);
  Eval.Clear();
  int nSteps = double(spec->GetNumberOfTimeSteps());
  int KernelStep = -1.0;
  int KernelSkip = 10;
  Config::Get("KernelEveryStep", KernelStep);
  Config::Get("KernelSkip", KernelSkip);
  if (KernelStep < 0)
    Exit("specify \"KernelEveryStep\" option");
  int countMax = int(floor(nSteps / double(KernelStep)));
  for (int count = KernelSkip; count < countMax; ++count) {
    Seis = 0.0;
    for (auto receiver = spec->begin(); receiver != spec->end(); ++receiver) {
      for (int i = 0; i < spec->GetNumberOfTimeSteps(); ++i) {
        double t = spec->TimeAtStep(i);
        FWIState Measure(0.0, 1);
        if (i == (KernelStep * count)) {
          Measure[0] = 1.0;
        }
        Seis.GetData().SetMeasurement(*receiver, t, Measure);
      }
    }
    AdjOp.CalculateAdjoint(Seis, Eval, nullptr);
    auto &Plot = mpp::plot("Kernel_" + to_string(count));
    Plot.AddData("Rho", Eval, 0);
    Plot.AddData("Vp", Eval, 1);
    if (numL > 0) {
      Plot.AddData("Taup", Eval, 2);
    }
    Plot.PlotFile();
  }
  return 0;
}

