//
// Created by cr on 25/01/24.
//

#ifndef FWI_PSEUDOHESSIANPRECONDITIONER_HPP
#define FWI_PSEUDOHESSIANPRECONDITIONER_HPP

#include <iomanip>
#include <utility>

#include "AAO/FWIIntegralOperator.hpp"
#include "AAO/FWILinearizedIntegralOperator.hpp"
#include "AAO/FWIRKOperator.hpp"
#include "AlgebraFwd.hpp"
#include "Assemble.hpp"
#include "Assembler/FWILinearizedAssembler.hpp"
#include "Assembler/FWIRhsAssemblerAdjoint.hpp"
#include "Assembler/FWIRhsAssemblerLinearized.hpp"
#include "BasicDoFs.hpp"
#include "BasicSolver.hpp"
#include "DGVAcousticAssembleBwd.hpp"
#include "DGVAcousticAssembleCombined.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "FWIAdjointCalculator.hpp"
#include "FWITaper.hpp"
#include "Helper/FWIDataContainerDomain.hpp"
#include "Helper/FWIDataContainerRange.hpp"
#include "Helper/FWIHelpers.hpp"
#include "Helper/FWIMppHelpers.hpp"
#include "Helper/GramianBlockInverse.hpp"
#include "Helper/NormCalculation.hpp"
#include "Helper/ReconstructionContainer.hpp"
#include "Helper/VectorsIntegrator.hpp"
#include "MeshesCreator.hpp"
#include "Observation/DGSeismogram.hpp"
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "Reduced/FWILinOpAdjRed.hpp"
#include "Reduced/FWILinOpRed.hpp"
#include "TimeIntegrator.hpp"
#include "VAcousticMatFromVec.hpp"
#include "lib1_math/basicalgebra/RMatrix.hpp"
#include "lib4_fem/elements/Element.hpp"

#include "m++.hpp"

class PseudoHessianPreconditioner {
  const std::shared_ptr<FWIVAcousticInversionProb> &problem_fwd;
  const EleCacheContainer &eleCacheContainer;
  const std::vector<unsigned int> &reconVec;
  const ReconstructionContainer &recon;
  double dt;
  std::unique_ptr<Matrix> MPC;

public:
  PseudoHessianPreconditioner() = delete;

  PseudoHessianPreconditioner(
      const std::shared_ptr<FWIVAcousticInversionProb> &problem_fwd_,
      const EleCacheContainer &eleCacheContainer_,
      const std::vector<unsigned int> &reconVec_,
      const ReconstructionContainer &recon_, double dt_)
      : problem_fwd(problem_fwd_),
        eleCacheContainer(eleCacheContainer_),
        reconVec(reconVec_),
        recon(recon_),
        dt(dt_) {}

  void assemble(const Vector &m, const Vector &m_update, const Vectors &u,
                double PreconRegu) {
    MPC = std::make_unique<Matrix>(m_update);
    *MPC = 0.0;
    Vectors u_fd(u);
    u_fd = 0.0;
    for (int i = 0; i < u.size() - 1; ++i) {
      u_fd[i] = u[i + 1];
      u_fd[i] -= u[i];
      u_fd[i] *= 1.0 / dt;
    }
    FWILinearizedAssembler fwiLinearizedAssembler(problem_fwd,
                                                  *eleCacheContainer.Get(), m);
    int Ninv = reconVec.size();
    Vector DiagonalCorrelation(0.0, m_update);
    Vector OffDiagonalCorrelation(0.0, m_update);
    Vector TmpMat(0.0, m_update);
    FWIAdjointCalculator adj(m, problem_fwd->getParametrization(),
                             *eleCacheContainer.Get(), recon,
                             problem_fwd->GetNumL(), dt, problem_fwd);

    for (int step = 0; step < u.size(); ++step) {
      double dt_ = (step == 0 || step == u.size() - 1) ? 0.5 * dt : dt;
      adj.PseudoHessianHelper(u[step], u_fd[step], u[step], TmpMat);
      for (const auto &elemIt : *eleCacheContainer.Get()) {
        const Cell &C = elemIt.second->GetCell();
        for (int param = 0; param < Ninv; param++) {
          DiagonalCorrelation(C(), param) +=
              dt_ * TmpMat(C(), reconVec[param]) * TmpMat(C(), reconVec[param]);
          for (int param2 = param + 1; param2 < Ninv; param2++) {
            OffDiagonalCorrelation(C(), param + param2 - 1) +=
                dt_ * TmpMat(C(), reconVec[param]) *
                TmpMat(C(), reconVec[param2]);
          }
        }
      }
      TmpMat = 0.0;
    }
    double maxi = 0;
    double mini = -infty;
    for (const auto &elemIt : *eleCacheContainer.Get()) {
      const Cell &C = elemIt.second->GetCell();
      for (int param = 0; param < Ninv; param++) {
        maxi = max(fabs(maxi), fabs(DiagonalCorrelation(C(), param)));
        mini = min(fabs(mini), fabs(DiagonalCorrelation(C(), param)));
        for (int param2 = param + 1; param2 < Ninv; param2++) {
          maxi = max(fabs(maxi),
                     fabs(OffDiagonalCorrelation(C(), param + param2 - 1)));
          mini = min(fabs(mini),
                     fabs(OffDiagonalCorrelation(C(), param + param2 - 1)));
        }
      }
    }
    double allmax = PPM->Max(maxi);
    DiagonalCorrelation /= allmax;
    OffDiagonalCorrelation /= allmax;
    double allmaxRegu = PreconRegu;
    double allmaxMatrix = -infty;
    for (const auto &elemIt : *eleCacheContainer.Get()) {
      SymRMatrix BlockMat(Ninv);
      const Cell &C = elemIt.second->GetCell();
      for (int p = 0; p < Ninv; p++) {
        BlockMat(p, p) = DiagonalCorrelation(C(), p);
        for (int p2 = p + 1; p2 < Ninv; p2++) {
          BlockMat(p, p2) = OffDiagonalCorrelation(C(), p + p2 - 1);
        }
      }
      for (int i = 0; i < Ninv; ++i) {
        BlockMat(i, i) += allmaxRegu;
      }
      BlockMat.Invert();
      DGRowEntries PHess_c(*MPC, C, C);
      for (int param = 0; param < Ninv; ++param) {
        allmaxMatrix = max(allmaxMatrix, BlockMat(param, param));
        PHess_c(reconVec[param], reconVec[param]) = BlockMat(param, param);
        for (int param2 = param + 1; param2 < Ninv; ++param2) {
          PHess_c(reconVec[param], reconVec[param2]) = BlockMat(param, param2);
          PHess_c(reconVec[param2], reconVec[param]) = BlockMat(param, param2);
        }
      }
    }
    double PPMALLMAXMATRIX = PPM->Max(allmaxMatrix);
    (*MPC) *= 1.0 / PPMALLMAXMATRIX;
    Vector PlotPC(m_update);
    for (const auto &elemIt : *eleCacheContainer.Get()) {
      const Cell &C = elemIt.second->GetCell();
      DGRowEntries PHess_c(*MPC, C, C);
      for (int param = 0; param < Ninv; ++param) {
        (PlotPC)(C(), param) = PHess_c(reconVec[param], reconVec[param]);
      }
    }
    (*MPC).CreateSparse();
  }
  void Apply(Vector &M_update) {
    Vector tmp = (*MPC) * M_update;
    M_update = tmp;
  }
};

#endif  // FWI_PSEUDOHESSIANPRECONDITIONER_HPP
