#ifndef _FWIHELPERS_H_
#define _FWIHELPERS_H_

#include "Cell.hpp"
#include "DGAcousticElement.hpp"
#include "DGDiscretization.hpp"
#include "Meshes.hpp"
#include "Parallel.hpp"
#include "Point.hpp"
#include "RVector.hpp"
#include "TimeIntegrator.hpp"
#include "Vertex.hpp"

#include "m++.hpp"
#define TIMEFUNCTION(CODE)                                  \
  mout << "Time before " << #CODE << " " << Date() << endl; \
  Date();                                                   \
  CODE;                                                     \
  mout << "Time after " << #CODE << " " << Date() << endl;



using EleCacheType = std::unordered_map<Point, DGAcousticElement *>;

class EleCacheContainer {
  std::shared_ptr<EleCacheType> ElemCells;
  std::shared_ptr<EleCacheType> ElemCellsOverlap;

public:
  EleCacheContainer(const VectorMatrixBase &Vec, int numL) {
    ElemCells = std::make_shared<EleCacheType>();
        ElemCellsOverlap = std::make_shared<EleCacheType>();
    for (cell c = Vec.cells(); c != Vec.cells_end(); ++c) {
      const Cell &C = *c;
      (*ElemCells)[C()] = new DGAcousticElement(Vec, C, numL);
    }
  }
  std::shared_ptr<EleCacheType> Get() const { return ElemCells; }

  ~EleCacheContainer() {
    for (auto elem : *ElemCells) {
      delete elem.second;
    }
  }
};

class FWIDiscretizationStack {
public:
  LevelPair lev{-1, -1};
  std::shared_ptr<Meshes> meshes = nullptr;
  std::shared_ptr<IDiscretization> mD = nullptr;
  std::shared_ptr<DGDiscretization> wD = nullptr;
  std::shared_ptr<Vector> mV = nullptr;
  std::shared_ptr<Vector> wV = nullptr;
  int deg;
  FWIDiscretizationStack() = default;

  FWIDiscretizationStack(const std::shared_ptr<Meshes> &meshes, int paramComps,
                         int stateComps, int deg_, LevelPair Lev)
      : lev(Lev),
        meshes(meshes),
        mD(std::make_shared<DGDiscretization>(*meshes, 0, paramComps)),
        wD(std::make_shared<DGDiscretization>(*meshes, deg_, stateComps)),
        mV(std::make_shared<Vector>(0.0, mD, lev)),
        wV(std::make_shared<Vector>(0.0, wD, lev)),
        deg(deg_) {}

  const Mesh &GetMesh() { return (*meshes)[lev]; }

  ~FWIDiscretizationStack() = default;
};
//class DiscStuff{
//  std::shared_ptr<FWIDiscretizationStack> Stack;
//  std::shared_ptr<EleCacheContainer> Container;
//public:
//  DiscStuff() = delete;
//  DiscStuff(const std::shared_ptr<Meshes> &meshes,
//            int paramComps,
//            int stateComps,
//            int deg_,
//            LevelPair Lev,
//            int numL)
//      : Stack(std::make_shared<FWIDiscretizationStack>
//          (meshes,paramComps,stateComps,deg_,Lev),
//        Container(*Stack.wV,numL)
//  {}
//};
struct RectangularConfig {
  double LeftBnd;
  double RightBnd;
  double BottomBnd;
  double TopBnd;
  int dimX;
  int dimY;
  int bndTop;
  int bndBottom;
  int bndLeft;
  int bndRight;
  std::string name;

  [[nodiscard]] std::vector<int> boundaries() const {
    return {bndLeft, bndRight, bndBottom, bndTop};
  }
};

std::shared_ptr<RectangularGeometry2D> getSharedGeom(
    const RectangularConfig &conf);
std::shared_ptr<Meshes> getFWIMesh();

class TimeConf {
  void readConfig() {
    std::string expfn{};
    Config::Get("ADDAXExperimentFile", expfn);
    if (!expfn.empty()) {
      json j;
      std::ifstream ifs(expfn + ".json");
      if (!ifs.good()) Exit("Failed to open file" + expfn);
      nlohmann::json Seismo = nlohmann::json::parse(ifs);
      std::vector<double> Times{};
      for (const auto &d : Seismo.at("Observation Specification").at("Times")) {
        Times.emplace_back(d);
      }
      t0 = Times[0];
      T = Times[1];
      dt = Times[2];
    } else {

      Config::Get("t0", t0);
      Config::Get("dt", dt);
      Config::Get("T", T);
    }
    Config::Get("rkorder", rkorder);
    TS = new TimeSeries(t0, T, dt);
    timeSteps = TS->Steps();
  }

  TimeSeries *TS = nullptr;

public:
 double t0 = 0;
 double dt = -1;
  double T = -1;
  int timeSteps = -1;
  int rkorder = 0;

  TimeConf() { readConfig(); }
};
class SourceConf {
public:
  std::string SignalName = "RickerDerivative";
  bool FrequencyFiltering = false;
  string FilteringType = "LowPass";
  double SignalFreq = 25.0;
  int FilteringOrder = 1;
  //    double LowPassFilterFreq = -1;
  vector<vector<double>> FrequencyLevels;
  int CurrentLevel = 0;
  int MaterialSmoothPasses = 0;
  double SignalDelay = 0;
  bool SaveObservedSeismogram = false;
  bool SaveFilteredSeismogram = false;
  double SignalCutoff = 0;
  bool SrcInversion = false;
  double SrcInversionKappa = 0.0;
  double SrcInversionEps = 0.0;
  std::vector<double> PreconRegu{1e-08};

  SourceConf();
  vector<double> GetFreq() const { return FrequencyLevels[CurrentLevel]; }
};

class WPConf {
public:
  int numL = 1;
  int dim = 2;
  std::string s_prob;
  int level = 0;
  int RealDataLevel;
  int InitialIterateLevel;
  TimeConf TConf;
  SourceConf SConf;

  WPConf();
};

auto GetFWITI(int rkorder) -> std::unique_ptr<TimeIntegrator>;

inline double Squared(double x) {
  return x * x;
}

inline std::string nr_str(const char *str, std::initializer_list<int> numbers) {
  for (auto number : numbers) {
    if (number >= 0) {
      std::stringstream ss;
      ss << str << "_";
      ss << number;
      return ss.str();
    }
  }
  return str;
}

// TODO: will not work for 1d and 2d....
inline std::istream &operator>>(std::istream &in, Point &P) {
  in >> P[0];
  in >> P[1];
  //    in >> P[2];
  return in;
}

inline std::istream &operator>>(std::istream &in, VectorField &v) {
  in >> v[0];
  in >> v[1];
  in >> v[2];
  return in;
}

// Copyright 2000 softSurfer, 2012 Dan Sunday
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.

// checks whether P2 is left of the line through P0,P1
inline Scalar isLeft(const Point &P0, const Point &P1, const Point &P2) {
  return ((P1[0] - P0[0]) * (P2[1] - P0[1]) -
          (P2[0] - P0[0]) * (P1[1] - P0[1]));
}
//===================================================================

// cn_PnPoly(): crossing number test for a point in a polygon
//      Return:  0 = outside, 1 = inside
// This code is patterned after [Franklin, 2000]
int cn_PnPoly(const Point &P, const cell &c);
//===================================================================

// wn_PnPoly(): winding number test for a point in a polygon
//      Return:  wn = the winding number (=0 only when P is outside)

int wn_PnPoly(const Point &P, const Cell &C);

// bool PointInCellDGTS_old(const Point &z, const Cell &C);
//
// bool PointInCellDGTS(const Point &z, const Cell &C);
//
// bool PointInCellDGTS(const Point &z, const cell &c);

inline double RickerDerivativeDelay(double f, double t, double delay,
                                    double cutoff) {
  if (t < delay || t > cutoff)
    return 0;
  double tau = Pi * f * ((t - delay) - 1.5 / f);
  return ((t - delay) - 1.5 / f) * exp(-tau * tau);
}

inline double RickerDelay(double f, double t, double delay, double cutoff) {
  if (t < delay || t > cutoff)
    return 0;
  double tau = Pi * f * ((t - delay) - 1.5 / f);
  return (1.0 - 2.0 * tau * tau) * exp(-tau * tau);
}

void fillQuadrature(RVector &c, RVector &b, int order);

inline void ProjectModel(const Vector &U2, Vector &UM, int components) {
  for (cell c = U2.cells(); c != U2.cells_end(); c++) {
    for (cell c2 = UM.cells(); c2 != UM.cells_end(); ++c2) {
      if (c.PointInCell(c2())) {
        for (int i = 0; i < components; ++i) {
          UM(c2(), i) = U2(c(), i);
        }
      }
    }
  }
}

inline void SmoothMaterial(Vector &M, int components, int passes = 1) {
  /* smoothing with kernel
 *    1
 *  1 2 1
 *    1
 */

  /* Applying this method with passes=2 is equivalent to convolving with the
   * following kernel
   *      1
   *    2 4 2
   *  1 4 8 4 1
   *    2 4 2
   *      1
   */

  for (int i = 0; i < passes; ++i) {
    Vector MM(M);
    MM.Accumulate();
    const Mesh &Me = M.GetMesh();
    M.Clear();
    for (cell c = M.cells(); c != M.cells_end(); ++c) {
      double counter = 0;
      // cell itself
      double fac = 2;
      for (int i = 0; i < components; ++i) {
        M(c(), i) += fac * MM(c(), i);
      }
      counter += fac;

      // face-neighbors
      for (int f = 0; f < c.Faces(); ++f) {
        Point N = M.neighbour_center(*c, f);
        if (N != Infty) {
          fac = 1;
          for (int i = 0; i < components; ++i) {
            M(c(), i) += fac * MM(N, i);
          }
          counter += fac;
        }
      }
      // normalize
      for (int i = 0; i < components; ++i) {
        M(c(), i) /= counter;
      }
    }
    M.MakeAdditive();
  }
}
#endif
