//
// Created by cr on 17.02.22.
//

#include "NormCalculation.hpp"

double NormCalculation::calculateTVNorm(Vector &Mat) {
  double norm = 0.0;
  for (const auto &elemIt: eleCache) {
    const auto &elem = *elemIt.second;
    const auto &C = elem.GetCell();
    double a = elem.Area();
    double diff = 0.0;
    for (int f = 0; f < C.Faces(); ++f) {
      Point N = Mat.neighbour_center(C, f);
      for (int direc = 0; direc < 2; ++direc) {
        if (N != Infty) {
          double h = N[direc] - C()[direc];
          if (h > Eps) {
            for (unsigned int k: reconlist) {
              double tmp = Mat(N, k);
              tmp -= Mat(C(), k);
              tmp *= 1.0 / h;
              tmp *= tmp;
              diff += tmp;
            }
          }
        }
      }
    }
    diff = sqrt(diff);
    diff *= a;
    norm += diff;
  }
  return PPM->Sum(norm);
}

void NormCalculation::calculateModelGradientCentral(const Vector &Mat, const string &direc_s,
                                                   Vector &grad) {
  int direc = -1;
  if (direc_s == "x") direc = 0;
  if (direc_s == "y") direc = 1;
  int paramComps = 3;
  grad = 0.0;
  for (cell c = grad.cells(); c != grad.cells_end(); ++c) {
    std::vector<double> diff(paramComps, 0.0);
    int left = -1;
    int right = -1;
    for (int f = 0; f < c.Faces(); ++f) {
      Point N = grad.neighbour_center(*c, f);
      double h = N[direc] - c()[direc];
      if (abs(h) > Eps && abs(h) < infty) {
        if (h > 0) right = f;
        if (h < 0) left = f;
      }
    }
    if (left > -1 && right > -1) {
      Point Nleft = grad.neighbour_center(*c, left);
      Point Nright = grad.neighbour_center(*c, right);
      double
          h = abs(Nleft[direc] - c()[direc]) + abs(Nright[direc] - c()[direc]);
      for (int i = 0; i < paramComps; ++i) {
        diff[i] = Mat(Nright, i);
        diff[i] -= Mat(Nleft, i);
        diff[i] *= (1.0 / h);
        grad(c(), i) = diff[i];
      }
    } else if (left > -1 && right == -1) {
      Point Nleft = grad.neighbour_center(*c, left);
      double h = abs(Nleft[direc] - c()[direc]);
      for (int i = 0; i < paramComps; ++i) {
        grad(c(), i) = Mat(c(), i);
        grad(c(), i) -= Mat(Nleft, i);
        grad(c(), i) *= (1.0 / h);
      }
    } else if (right > -1 && left == -1) {
      Point Nright = grad.neighbour_center(*c, right);
      double h = abs(Nright[direc] - c()[direc]);
      for (int i = 0; i < paramComps; ++i) {
        grad(c(), i) = Mat(Nright, i);
        grad(c(), i) -= Mat(c(), i);
        grad(c(), i) *= (1.0 / h);
      }
    }
  }
}

Vector NormCalculation::calculateTVNormGradient(const Vector &Mat, Vector &GradX, Vector &GradY,
                                               int order) const {
  if (order == 1) {
    calculateModelGradient(Mat, "x", GradX);
    calculateModelGradient(Mat, "y", GradY);
  } else if (order == 2) {
    calculateModelGradientCentral(Mat, "x", GradX);
    calculateModelGradientCentral(Mat, "y", GradY);
  } else {Exit("wrong order")}
  Vector TMP(GradX);
  TMP = 0;
  for (cell c = GradX.cells(); c != GradX.cells_end(); ++c) {
    for (unsigned int k: reconlist) {
      double tmp = GradX(c(), k);
      tmp *= tmp;
      TMP(c(), k) = tmp;
      tmp = GradY(c(), k);
      tmp *= tmp;
      TMP(c(), k) += tmp;
      TMP(c(), k) = sqrt(TMP(c(), k));
    }
  }
  return TMP;
}

void NormCalculation::calculateModelGradient(const Vector &Mat, const string &direc_s,
                                            Vector &grad) const {
  int direc = -1;
  if (direc_s == "x") direc = 0;
  if (direc_s == "y") direc = 1;
  const Mesh &M = grad.GetMesh();
  grad = 0.0;
  for (cell c = grad.cells(); c != grad.cells_end(); ++c) {
    for (int f = 0; f < c.Faces(); ++f) {
      Point N = grad.neighbour_center(*c, f);
      if (N != Infty) {
        double h = N[direc] - c()[direc];
        if (abs(h) > Eps) {
          for (unsigned int k: reconlist) {
            grad(c(), k) = Mat(N, k);
            grad(c(), k) -= Mat(c(), k);
            grad(c(), k) *= 1.0 / h;
          }
        }
      }
    }
  }
}

double nc::CellWiseStepProduct(const Vector &U, const Vector &V, const DGAcousticElement &elem,
                               double dt) {
  double e = 0.0;
  for (int q = 0; q < elem.nQ(); ++q) {
    double w = elem.QWeight(q);
    e += w * dt * elem.Velocity(q, U) * elem.Velocity(q, V);
    for (int n = 0; n < elem.p_dimension(); ++n) {
      e += w * dt * elem.Pressure_i(q, U, n) * elem.Pressure_i(q, V, n);
    }
  }
  return e;
}

double nc::CellWiseTimeProduct(const Vectors &U, const Vectors &V, const DGAcousticElement &elem,
                               double dt) {
  double skp = 0.0;
  for (int l = 0; l < U.size(); ++l) {
    if (l == 0 || l == U.size() - 1) {
      skp += CellWiseStepProduct(U[l], V[l], elem, 0.5 * dt);
    } else skp += CellWiseStepProduct(U[l], V[l], elem, 1.0 * dt);
  }
  return skp;
}

double nc::L2tScalar(const Vector &u, const Vector &v,
                     const std::shared_ptr<EleCacheType> &eleCache, double dt) {
  Scalar e = 0;
  for (const auto &elemIt: *eleCache) {
    e += CellWiseStepProduct(u, v, *elemIt.second, dt);
  }
  return PPM->Sum(e);
}

double nc::L2TScalar(const Vectors &u, const Vectors &v,
                     const std::shared_ptr<EleCacheType> &eleCache, double dt) {
  double l2t = 0.0;
  for (int tsteps = 0; tsteps < u.size(); ++tsteps) {
    if (tsteps == 0 || tsteps == u.size() - 1) {
      l2t += 0.5 * L2tScalar(u[tsteps], v[tsteps], eleCache, dt);
    } else {
      l2t += L2tScalar(u[tsteps], v[tsteps], eleCache, dt);
    }
  }
  return l2t;
}

double nc::L2TNorm(const Vectors &u,
                   const std::shared_ptr<EleCacheType> &eleCache, double dt) {
  return sqrt(L2TScalar(u, u, eleCache, dt));
}

double nc::L2MaterialScalar(const Vector &u, const Vector &v,
                            const std::shared_ptr<EleCacheType> &eleCache,
                            const std::vector<unsigned int> &reconlist) {
  double l2 = 0.0;
  for (const auto &elemIt: *eleCache) {
    const auto& elem = *elemIt.second;
    const auto& C = elem.GetCell();
    double a = elem.Area();
    row ru = u.find_row(C());
    row rv = v.find_row(C());
    for (unsigned int k: reconlist) {
      double para_u = u(ru, k);
      double para_v = v(rv, k);
      l2 += para_u * para_v * a;
    }
  }
  return PPM->Sum(l2);
}

double nc::L2MaterialNorm(const Vector &u,
                          const std::shared_ptr<EleCacheType> &eleCache,
                          const std::vector<unsigned int> &reconlist) {
  return sqrt(L2MaterialScalar(u, u, eleCache, reconlist));
}
