#ifndef _DGVISCOACOUSTICASSEMBLEMATRIX_HPP_
#define _DGVISCOACOUSTICASSEMBLEMATRIX_HPP_

#include "DGDiscretization.hpp"
#include "DGElement.hpp"
#include "DGViscoAcousticProblem.hpp"
#include "LagrangeDiscretization.hpp"
#include "Plotting.hpp"
#include "ctools.hpp"
#include "DGVAcousticPlotter.hpp"
#include "DGVAcousticMatrixAssembler.hpp"
#include "DGVAcousticRHSAssembler.hpp"
#include "Assemble.hpp"

class DGAssemble : public ILinearTimeAssemble {
  bool useCollectTweak = false;
protected:
  std::shared_ptr<ProblemVAcoustic> problem;
  bool upwind = true;
  const std::unordered_map<Point, DGAcousticElement *> *EleCache;
  bool clearCache;
  int plotting = 0;
  DGVAcousticPlotter plotter;
  DGVAcousticMatrixAssembler acousticMatrixAssembler;
  std::shared_ptr<DGVAcousticRHSAssembler> rhsAssembler;
public:
  DGAssemble(std::shared_ptr<ProblemVAcoustic> P, bool fwd, int printRead, double t0,double T,double dt)
      : ILinearTimeAssemble(t0,T,dt),
        problem(std::move(P)),
        EleCache(nullptr),
        clearCache(false),
        plotter(problem->NumDamping()),
        acousticMatrixAssembler(
            std::make_unique<DGVAcousticMaterialFromProblem>(
                *problem),
            fwd),
        rhsAssembler(
            std::make_shared<DGVAcousticRHSAssemblerDefault>(
                *problem)) {
    Config::Get("UseCollectTweak", useCollectTweak);
    Config::Get("plot_tStep", plotting, true);
  }

  void Initialize(Vector &u) const override {
    rhsAssembler->Initialize(u);
  };

  ~DGAssemble() {
    if (clearCache) {
      for (auto &it: *EleCache) {
        delete it.second;
      }
    }
  }

  void SetElementCache(const std::unordered_map<Point, DGAcousticElement *> *_EleCache);

  void PrintInfo(const Vector &u) {
    Exit("fix printinfo");
    vout(1) << " Mesh            " << problem->MeshName() << endl
            << " Problem         " << problem->Name() << endl << endl;
  }

  DGAcousticElement *getElem(const VectorMatrixBase &base, const Cell &C) const;

  void destroyElement(DGAcousticElement *elem) const;

  const char *Name() const override { return "DGAssembleMatrix"; }

  void MassMatrix(Matrix &M) const override;

  void DampingMatrix(Matrix &D) const;

  virtual void FluxMatrix(Matrix &A) const;

  void FluxMatrix(Matrix &A, bool, bool) const;

  void SystemMatrix(Matrix &S) const override;

  virtual void FluxMatrixOnly(Matrix &A) const;

  bool RHS(double t) const { return problem->RHS(t); }

  virtual void RHS(double t, Vector &b) const override;

  void PlottingExact(const Vector &u, int step, double t) const;

  [[nodiscard]] double Norm(const Vector &u) const { return sqrt(Energy(u)); }

  void VtkPlotting(const Vector &u, int step) const;

  void SetInitialValue(Vector &u) override { get_u_int_assemble(0, u); }

  void SetInitialValueOld(Vector &u) const { get_u_int_assemble(0, u); }

  void Interpolate(const Vector &u, Vector &U) const;

  void FinishTimeStep(const Vector &u) {
    if (plotting <= 0 || Step() % plotting != 0) return;
    PrintStep(Time(), u, Step());
    VtkPlotting(u, Step());
  }

  void InitialPlotting(const Vector &u) const;

  void FinishStep(Vector &u, double t, double dt, int step,
                  int plot_step = 1) const;

  void InitialStep(double t, Vector &u, bool print = true) const;

  void PrintStep(double t, const Vector &u, int step) const;

  virtual double L2Error(double t, const Vector &u) const;

  double Energy(const Vector &u) const;

  double Energy_local(const Vector &u,
                      DGAcousticElement &elem, const cell &c) const;

  virtual void get_u_int_assemble(double t, Vector &u) const;
};

#endif