#ifndef _DGVISCOACOUSTIC_H_
#define _DGVISCOACOUSTIC_H_

#include <fstream>

#include <LagrangeDiscretization.hpp>
#include <Plotting.hpp>

#include "DGDiscretization.hpp"
#include "DGElement.hpp"
#include "Helper/FWIHelpers.hpp" // TODO: move PointInCellDGTS into M++-Library
#include "ctools.hpp"

class ProblemVAcoustic {
protected:
  std::string meshname;
  int dim;
  int numL;
  int nrPointSrc;
  std::vector<Point> sourceLocations;

public:
  ProblemVAcoustic(int _numL = -1, int d = 3)
      : meshname("none"), dim(d), numL(_numL), nrPointSrc(0) {
    Config::Get("Mesh", meshname, -1);
  }

  virtual ~ProblemVAcoustic() {}

  int Dim() const { return dim; }

  int NumDamping() const { return numL; }

  virtual int getNrPointSrc() const { return nrPointSrc; }

  virtual Point getPointSrcLoc(int nr) const { return sourceLocations[nr]; }

  virtual bool HasExactSolution() const { return false; }

  virtual bool RHS(double) const { return false; }

  virtual void Force(VectorField &V_rhs, RVector &Force, double t, const Cell &C,
                     const Point &z) const {
    V_rhs = ForceV(t, C, z);
    for (int n = 0; n < Force.size(); ++n) {
      Force[n] = ForceP_i(t, C, z, n);
    }
  }

  virtual double ForceP_i(double, const Cell &, const Point &, int i) const {
    Exit("specify force in inherited class") return 0;
  }

  virtual VectorField ForceV(double, const Cell &, const Point &) const {
    Exit("specify force in inherited class")

        return zero;
  }

  virtual double Dirichlet_i(double t, const Cell &C, const Point &z, int i) const {
    return ut(t, C, z, dim + i);
  }

  virtual double Neumann(double t, const Cell &C, const Point z, const VectorField &N) const {
    VectorField V(ut(t, C, z, 0), ut(t, C, z, 1));
    return V * N;
  }

  virtual double ut(double t, const Point &, int) const {
    Exit("Not Implemented!");
    return 0;
  }

  virtual double ut(double t, const Cell &C, const Point &x, int i) const { return ut(t, x, i); }

  virtual double u0(const Point &x, int i) const { return ut(0, x, i); }

  virtual double u0(const Cell &C, const Point &x, int i) const { return ut(0, C, x, i); }

  virtual double Kappa(const Cell &C, const Point &x) const {
    double kappa = 0;
    for (int i = 0; i <= numL; ++i) {
      kappa += Kappa_i(C, x, i);
    }
    return kappa;
  }

  virtual double Kappa_i(const Cell &C, const Point &x, int i) const {
    Exit("not implemented");
    return 0.0;
  }

  virtual double invKappa_i(const Cell &C, const Point &x, int i) const {
    return 1.0 / Kappa_i(C, x, i);
  }

  virtual double Velocity(const Point &x) const {
    Exit("Velocity no modelparameter in this case.");
    return 0.0;
  }

  virtual double Tau_i(const Cell &c, const Point &x, int i) const { return Tau_i(c, x, i); }

  virtual double invTau_i(const Cell &C, const Point &x, int i) const {
    return 1.0 / Tau_i(C, x, i);
  }

  virtual double Rho(const Cell &C, const Point &x) const {
    Exit("notimplemented");
    return 0.0;
  }

  virtual const char *Name() const {
    Exit("not implemented");
    return "";
  }

  virtual const char *MeshName() const { return meshname.c_str(); }
  virtual double alpha(const Cell &, const Point &) const {
    Exit("not implemented here!");
    return 0.;
  }
  virtual double alpha1() const {
    Exit("not implemented here!");
    return 0.;
  }
  virtual double alpha2() const {
    Exit("not implemented here!");
    return 0.;
  }
};

class VAcoustic_Benchmark_Constant1 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;

public:
  VAcoustic_Benchmark_Constant1()
      : ProblemVAcoustic(1, 2), rho(1.0), kappa1(1.0), kappa0(1.0), tau1(1.0) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const { return zero; }

  double ForceP_i(double t, const cell &c, const Point &x, int i, int q = -1) const {
    switch (i) {
    case 0:
      return 0.0;
    case 1:
      return 0.0;
    default:
      Exit("Wrong call")
    }
    return 0.0;
  }

  const char *Name() const { return "Constant p0=3 ,p1=0, v = (1,2)"; }

  double ut(double t, const Point &x, int i) const {
    switch (i) {
    case 0:
      return 1;
    case 1:
      return 2;
    case 2:
      return 3;
    case 3:
      return 0;
    default:
      return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const { return rho; }

  double Tau_i(const Cell &c, const Point &x, int i) const { return tau1; }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
    case 0:
      return kappa0;
    case 1:
      return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Benchmark_Linear1 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;

public:
  VAcoustic_Benchmark_Linear1() : ProblemVAcoustic(1, 2), tau1(1), kappa0(1), kappa1(1), rho(1) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const Cell &C, const Point &z, int q = -1) const { return zero; }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
    case 0:
      return 0.0;
    case 1:
      return (t + Tau_i(C, x, i)) / (Kappa_i(C, x, i) * Tau_i(C, x, i));
    default:
      Exit("Wrong call");
      return 0.0;
    }
  }

  const char *Name() const { return "Linear1 p0=0 ,p1=t, v =(0,0)"; }

  double ut(double t, const Point &x, int i) const {
    switch (i) {
    case 0:
      return 0;
    case 1:
      return 0;
    case 2:
      return 0;
    case 3:
      return t;
    default:
      return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const { return rho; }

  double Tau_i(const Cell &c, const Point &x, int i) const { return tau1; }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
    case 0:
      return kappa0;
    case 1:
      return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Benchmark_Linear2 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;

public:
  VAcoustic_Benchmark_Linear2() : ProblemVAcoustic(1, 2), tau1(1), kappa0(1), kappa1(1) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const Cell &C, const Point &z, int q = -1) const { return zero; }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
    case 0:
      return 0.0;
    case 1:
      return -(1.0 / Kappa(C, x));
    default:
      Exit("Wrong call")
    }
    return 0.0;
  }

  const char *Name() const { return "Linear2 p0=t ,p1=0, v = (0,0)"; }

  double ut(double t, const Cell &C, const Point &x, int i) const {
    switch (i) {
    case 0:
      return x[0] / Kappa_i(C, x, 0);
    case 1:
      return 0;
    case 2:
      return t;
    case 3:
      return 0;
    default:
      return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const { return rho; }

  double Tau_i(const Cell &c, const Point &x, int i) const { return tau1; }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
    case 0:
      return kappa0;
    case 1:
      return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Benchmark_Linear3 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;

public:
  VAcoustic_Benchmark_Linear3() : ProblemVAcoustic(1, 2), tau1(1), kappa0(1), kappa1(1), rho(1.0) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const Cell &C, const Point &z, int q = -1) const { return zero; }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
    case 0:
      return 0.0;
    case 1:
      return -(1.0 / Kappa_i(C, x, 0)) + (t + Tau_i(C, x, 1)) / (Kappa_i(C, x, 0) * Tau_i(C, x, 1));
    default:
      Exit("Wrong call")
    }
    return 0.0;
  }

  const char *Name() const { return "Constant p0=t ,p1=t, v = (x/kappa0,0)"; }

  double ut(double t, const Cell &C, const Point &x, int i) const {
    switch (i) {
    case 0:
      return x[0] / Kappa_i(C, x, 0);
    case 1:
      return 0;
    case 2:
      return t;
    case 3:
      return t;
    default:
      return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const { return rho; }

  double Tau_i(const Cell &c, const Point &x, int i) const { return tau1; }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
    case 0:
      return kappa0;
    case 1:
      return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcousticSinCos2d_RHS : public ProblemVAcoustic {
  double kappa1;
  double tau1;
  double rho;
  double omega;
  double kappa0;

public:
  VAcousticSinCos2d_RHS() : ProblemVAcoustic(1, 2), kappa0(1), rho(1), tau1(1), kappa1(1) {
    omega = 1.0 / 4.0;
    Config::Get("omega", omega);
    omega *= 2 * M_PI;
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
    Config::Get("kappa0", kappa0);
    Config::Get("tau1", tau1);
    mout << "Kappa1: " << kappa1 << " Kappa0: " << kappa0 << " Rho: " << rho << " tau1: " << tau1
         << endl;
  }

  const char *Name() const { return "VAcousticSinCos2d_RHS p,q = TODO"; }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  double ut(double t, const Point &x, int i) const {
    switch (i) {
    case 0:
      return Velocity(t, x)[0];
    case 1:
      return Velocity(t, x)[1];
    case 2:
      return Pressure0(t, x);
    case 3:
      return Pressure1(t, x);
    }
    Exit("Not implemented");
    return 0.0;
  }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const { return zero; }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
    case 0:
      return ForceP0(t, x);
    case 1:
      return ForceP1(t, C, x);
    default:
      Exit("Wrong number of Kappas");
    }
    return 0.0;
  }

  double Pressure1(double t, const Point &x) const {
    return 0.7 * sin(omega * x[0]) * sin(omega * x[1]);
  }

  double Pressure0(double t, const Point &x) const {
    return 0.3 * sin(omega * x[0]) * sin(omega * x[1]);
  }

  VectorField Velocity(double t, const Point &x) const {
    return VectorField((t - 1) * omega * sin(omega * x[1]) * cos(omega * x[0]),
                       (t - 1) * omega * sin(omega * x[0]) * cos(omega * x[1]), 0.0);
  }

  int Deg() const { return 1000; }

  double ForceP0(double t, const Point &x) const {
    return 2 * (t - 1) * omega * omega * sin(omega * x[0]) * sin(omega * x[1]);
  }

  double ForceP1(double t, const Cell &C, const Point &x) const {
    return 2 * (t - 1) * omega * omega * sin(omega * x[0]) * sin(omega * x[1]) +
           1.0 / (Tau_i(C, x, 1) * Kappa_i(C, x, 1)) * Pressure1(t, x);
  }

  int SourceDeg() const { return 1000; }

  double Rho(const Cell &, const Point &) const { return rho; }

  double Tau_i(const Cell &, const Point &, int) const { return tau1; }

  double Kappa_i(const Cell &, const Point &, int i) const {
    switch (i) {
    case 0:
      return kappa0;
    case 1:
      return kappa1;
    default:
      Exit("wrong kappa!")
    }
    return 0.0;
  }
};

inline ProblemVAcoustic *GetProblemVAcoustic(const std::string &name) {
  int dim = 2;
  Config::Get("dim", dim);
  if (name == "Benchmark_Linear1") return new VAcoustic_Benchmark_Linear1();
  if (name == "Benchmark_Linear2") return new VAcoustic_Benchmark_Linear2();
  if (name == "Benchmark_Linear3") return new VAcoustic_Benchmark_Linear3();
  if (name == "Benchmark_Constant1") return new VAcoustic_Benchmark_Constant1();
  if (name == "VAcousticSinCos2d_RHS") return new VAcousticSinCos2d_RHS();
  else Exit("Error: no wave found!");
  return nullptr;
}

#endif
