import numpy as np

from numpy import pi, abs, sqrt, exp, real, conj, nan_to_num, heaviside
from numpy.fft import fft, ifft, fftfreq, fftshift, ifftshift, irfft

import scipy as sp
import scipy.signal
import scipy.special
from scipy.special import hankel1, hankel2

import matplotlib.pyplot as plt
import sys
import copy

from seismogram_lib_conv import FWISeismogram, plot_phase_spectrum, plot_magnitude_spectrum
import multiprocessing as mp


# --------------------------------
# time domain

def H(x):
    return heaviside(x, 0.0)

def G(t, r, shift=0.0, c=1):
    t = t - shift
    t[t < 0] = 0
    t2 = t ** 2 - r ** 2 / c ** 2
    #print(t2)
    regularization = 1e-9
    v = np.empty(t2.shape)
    isRelevant = t2 > regularization
    v[t2 <= regularization] = 0.0
    #Evaluation = t[isRelevant] - abs(r) / c
   # print(min(t2[isRelevant]),min(Evaluation))
    #for i in range(len(Evaluation)):
        #Evaluation[i] = max(1e-32, Evaluation[i])
    v[isRelevant] = 1.0 / sqrt(t2[isRelevant])
    return v / (2.0 * pi)

def dt_gauss(t, sigma=1):
    return -2*t*exp(-(t/sigma)**2/2) / (sigma*sqrt(2*pi))

def ricker(t, freq): 
    a = (pi**2)*(freq**2)
    return (1.-2.*a*t**2)*exp(-a*t**2)

def dt_ricker(t, freq):
    a = (pi**2)*(freq**2)
    return (4*a**2*t**3 - 6*a*t) * exp(-a*t**2)


def get_pressure_trace(conf, receiver_pos):
    dt = conf["dt"]
    c=conf["c"]
    r = np.linalg.norm(conf["source_pos"]-receiver_pos)
    start = r/c
    t_solution = np.arange(*conf["time_interval"], dt)
    t_source = np.arange(*conf["source_time_interval"], dt)
    t_padded = np.arange(conf["time_interval"][0]+conf["source_time_interval"][0], 
                         conf["time_interval"][1]+conf["source_time_interval"][1], dt)

    shift = conf["source_t_offset"]
    green = G(t_padded,r, shift=shift, c=conf["c"])
   # print(np.sqrt(np.dot(green, green)))
    source = conf["source_func"](t_source,freq=conf["f"])
   # print(np.sqrt(np.dot(source, source)))
    print(np.max(green))
    #print("AAA", fft(green)*(dt/100))
    
    solution = sp.signal.oaconvolv(green, source, mode="valid")*(dt)
    solution = 1.0/conf["c"]**2 * solution[::]*conf["source_factor"]
    #5solution= np.cumsum(solution)
    #plt.plot(t_padded,green); plt.show(); sys.exit()
    #plt.plot(t_source,source); plt.show(); sys.exit()
    #plt.plot(solution); plt.show(); sys.exit()
    
    return t_solution, solution

# --------------------------------
# Frequency domain

def ricker_freq(omega, freq):
    omega_p = 2*pi*freq
    omega2 = omega**2
    return omega2/(2. * pi**2.5 * freq**3) * exp(-omega2/omega_p**2)

def dt_ricker_freq(omega,freq):
    r = ricker_freq(omega,freq)
    return 1j*omega * r

def delta_freq(omega, freq):
    omega_p = 2*pi*freq
    return exp(-omega_p)

def dt_delta_freq(omega,freq):
    r = delta_freq(omega,freq)
    return 1j*omega * r


def G_freq(omega, r, c=1.):
    # Note: the magic constant 0.2781899 is obtained experimetally by transforming the time-domain Greens function
    # Calzone says that setting the 0-th order term to 0 does not change much since the corresponding part in the source is usually small

    #return np.hstack(([0.2781899], 1j * pi * hankel1(0, arg[1:]) / (4*pi)))
        
    arg = -(r/c)*omega
    N = (len(omega)-1)//2
    arg_pos = arg[1:N+1]
    G_pos = -1j * pi * nan_to_num(hankel1(0, arg_pos)) / (4 * pi)  # see Carcione
    
    G = np.hstack(([0.2781899], G_pos, G_pos.conj()[::-1]))
    return G
    
    #return -1j * pi * nan_to_num(hankel2(0, arg)) / (4*pi)


def get_pressure_trace_freq(conf, dim, receiver_pos):
    dt = conf["dt"]
    r = np.linalg.norm(conf["source_pos"]-receiver_pos)
    
    t_min, t_max = conf["time_interval"]
    t_solution = np.linspace(t_min,t_max, int(round((t_max-t_min)/dt))+1)
    N, fac = len(t_solution), 7
    fft_omegas = (2.*pi) * fftfreq(N*fac, dt/fac)
    fft_green = G_freq(fft_omegas, r, c=conf["c"])
    fft_source = exp(-1.j*t_min * fft_omegas * conf["source_t_offset"]) * conf["rho"] * dt_ricker_freq(fft_omegas,conf["f"])
    
    # diract delta rhs
    #fft_source = exp(1j*t_min * fft_omegas) * conf["rho"] * 1
    
    
    #fft_source = exp(1j*t_min * fft_omegas) * conf["rho"] \
                    #* dt_delta_freq(fft_omegas,conf["f"])
                
    #fft_source *= exp(-1.j*conf["source_t_offset"] * fft_omegas) # time offset of source
    
    fft_solution = fft_green * fft_source 
    solution_from_fft = (ifft(fft_solution)) / (dt/fac)
    #solution_from_fft = (irfft(fft_solution, n=N_pos*2+1)) / (dt/fac)
    
    solution_from_fft = solution_from_fft[::fac]

    #plt.plot(t_padded,green); plt.show(); sys.exit()
    #plt.plot(t_source,source); plt.show(); sys.exit()
    #plt.plot(solution); plt.show(); sys.exit()
    
    return t_solution, solution_from_fft

# --------------------------------
def get_pressure_trace_write(conf,ts,i):
    dt = conf["dt"]
    dtfac = 100
    receiver_pos = np.array([ts.receiver.x, ts.receiver.y])
    r = np.linalg.norm(conf["source_pos"]-receiver_pos)
    c=conf["c"]
    start = r/c
    #print( start % (dt/dtfac))
    #print(*conf["time_interval"]+start % (dt/dtfac) )
    t_solution = np.arange(conf["time_interval"][0],conf["time_interval"][1]+ dt, dt )
    t_solution_shift = np.arange(conf["time_interval"][0]+ start % (dt/dtfac),conf["time_interval"][1]+ start % (dt/dtfac) +dt/dtfac , dt/dtfac)

    t_source = np.arange(*conf["source_time_interval"]+start % (dt/dtfac), dt/dtfac)
    t_padded = np.arange(conf["time_interval"][0]+conf["source_time_interval"][0] + start % (dt/dtfac) ,conf["time_interval"][1]+conf["source_time_interval"][1] + start % (dt/dtfac), dt/dtfac)
    #print(t_padded)
    shift = conf["source_t_offset"]
    green = G(t_padded,r, shift=shift, c=conf["c"])
    source = conf["source_func"](t_source,freq=conf["f"])
    solution = sp.signal.oaconvolve(green, source, mode="valid")*(dt/dtfac)
    #print("t_source",len(t_source))
    #print("green",len(green))
    #print("t_padded",len(t_padded))
    #print("solution",solution.shape)
    #print("t_solution_shift",t_solution_shift.shape)
    #print(t_solution_shift)
    #print(t_solution_shift)
    interpol = sp.interpolate.interp1d(t_solution_shift, solution)
    solution_new = np.hstack((np.array([0.0]), 1.0/conf["c"]**2 * interpol(t_solution[1::]) * conf["source_factor"]))
    return i,solution_new

def create_pressure_seismogram(seismogram, conf,dim):

    conf = conf.copy()
    spec = seismogram.spec
    conf["time_interval"] = [spec.t_start, spec.t_end]
    conf["dt"] = spec.t_delta
    nrproc = 128
    new_s = copy.deepcopy(seismogram)
    itera = [(conf,ts,i) for i,ts in enumerate(new_s.timeseries)]
    with mp.Pool(nrproc) as pool:
        result = pool.starmap(get_pressure_trace_write, itera)
    for i,sol in result:
        new_s.timeseries[i].states[0:len(sol),0] = sol.copy()
    #rec = np.array(len(new_s.timeseries),2)
    #for ts in new_s.timeseries:
        #rec
   # for ts in new_s.timeseries:
    #    receiver_pos = np.array([ts.receiver.x, ts.receiver.y])
      #  t, p = get_pressure_trace(conf, receiver_pos=receiver_pos)
     #   #t, p = get_pressure_trace_freq(conf,dim, receiver_pos=receiver_pos)
       # print(min(p), max(p), min(ts.states), max(ts.states))
       # ts.states[0:len(p),0] = p.copy()
    new_s.label = seismogram.label + 'exact'
    return new_s



if __name__ == "__main__":
    import sys
    import argparse
    parser = argparse.ArgumentParser(description='AcousticSol')
    parser.add_argument("--showPlot",type=int,default=1)
    parser.add_argument("--SrcLoc", type=float,nargs=2)
    parser.add_argument("-c", type=float, default=3500)
    parser.add_argument("-f", type=float, default= 25.0)
    parser.add_argument("-s", type=str, default="")
    parser.add_argument
    parser.add_argument("--SrcFactor", type=float, default=1.0)
    parser.add_argument("--ExactSeismoName", type=str, default="")


    args = parser.parse_args()
    if args.s =="":
        exit("Set SeismogramName!!")
    dim = 2
    conf = {}
    if dim == 1:
        comp = 1
        conf["c"] = 5.
        conf["rho"] = 1.
        conf["f"] = 25.0
        conf["source_pos"] = np.array([args.SrcLoc[0], args.SrcLoc[1]]),
        conf["source_t_offset"] = 1.5/conf["f"]
        conf["source_func"] = dt_ricker
        conf["source_time_interval"] = [-2./conf["f"],2./conf["f"]]
    if dim == 2:
        comp = 2
        conf["c"] = args.c
        conf["rho"] = 2000.0
        conf["f"] = args.f
        conf["source_pos"] = np.array([args.SrcLoc[0],args.SrcLoc[1]]),
        conf["source_t_offset"] = 1.5/conf["f"]
        conf["source_func"] = ricker
        conf["source_time_interval"] = [-2./conf["f"],2./conf["f"]]
        conf["source_factor"] = args.SrcFactor


    s = FWISeismogram.readFromFile(args.s).extractComponent(0)

    s_sol = create_pressure_seismogram(s, conf, dim)

    d_min, d_max = np.inf, -np.inf
    for seis in [s, s_sol]:
        d_min = min(d_min, seis.min())
        d_max = max(d_max, seis.max())

    plot_spectra = 0
    if plot_spectra:
        print("WARNING: hardcoded stuff....")
    
        plot_magnitude_spectrum([s_sol], 0, figsize=(6,4), xlim=(0,270), ylim=(0,2000), filename="exact_amplitude.pdf")
    
        plot_phase_spectrum([s_sol], 0, figsize=(6,4), xlim=(0,250), ylim=(-350,0), filename="exact_phase.pdf")
    
    else:

        if args.ExactSeismoName != "":
            s_sol.writeToFile(args.ExactSeismoName)
        #diff = s_sol - s
        #print("Rel. L2-Difference: ", diff.norm2()/s_sol.norm2())
        if args.showPlot == 1:
            figsize=(3.5,10)
            figure = plt.figure(figsize=figsize)
        #data_factor = "normalized"
            data_factor = 1
            s.plot(data_factor=data_factor, data_range=(d_min,d_max),
                figure=figure, show=False, fill_below=False, r_step=1)
            s_sol.plot(data_factor=data_factor, data_range=(d_min,d_max),
                figure=figure, show=False, fill_below=False,
                linestyle="--", r_step=1)
            plt.show()
